(* $Id: godi_make.ml 808 2008-06-12 21:47:00Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

(* NEW:
 * ${var:n} convert filename ${var} to native format (windows)
 *)

(* TODO:
 *
 * LATER:
 * - ${MAKEFLAGS} simulation, ${.MAKEOVERRIDES}
 *   ${MAKEFLAGS}=${MFLAGS}, 
 * - set $PWD and $MAKEFLAGS when executing commands
 * - ${MAKEOBJDIR}, ${MAKEOBJDIRPREFIX}
 * - ${MAKE_PRINT_VAR_ON_ERROR}
 * - Special sources:
 *   .EXEC               rule_exec_flag
 *    (same as "!" dependency)
 *   .IGNORE             rule_ignore_flag
 *   .INVISIBLE          rule_invisible_flag
 *    (not part of $(OODATE))
 *   .JOIN               rule_join_flag
 *    (are purely logical node) CHECK
 *    difference to phony: phony nodes are always out of date, whereas join
 *    nodes are only out of date if any of the children is out of date
 *   .MADE               rule_made_flag
 *   .MAKE               rule_force_exec_flag
 *   .NOPATH ignore
 *   .NOTMAIN            does not set ctx.default_target
 *   .OPTIONAL           rule_optional_flag
 *   .PRECIOUS           rule_precious_flag ctx.global_precious_flag
 *   .SILENT             rule_silent_flag ctx.global_silent_flag
 *   .USE error
 *   .USEBEFORE error
 *   .WAIT ignore
 * - Special targets:
 *   .BEGIN              propagate to "::" style
 *   .DEFAULT error
 *   .END                propagate to "::" style
 *   .IGNORE             rule_ignore_flag global_ignore_flag
 *   .INCLUDES ignore
 *   .INTERRUPT          propagate to "::" style
 *   .LIBS ignore
 *   .MAIN               set ctx.default_target
 *   .MAKEFLAGS          set ${.MAKEFLAGS}
 *   or .MFLAGS
 *   .NOPATH ignore
 *   .NOTPARALLEL ignore
 *   .NO_PARALLEL ignore
 *   .NULL ignore
 *   .ORDER              CHECK
 *   .PATH ignore
 *   .PHONY              rule_phony_flag
 *   .POSIX ignore
 *   .PRECIOUS           rule_precious_flag
 *   .SHELL ignore
 *   .SILENT             rule_silent_flag
 *   .SINGLESHELL ignore
 *      create a shell for each command
 *   .SUFFIXES           ctx.suffixes
 *                       ctx.suffix_rules
 *   .WAIT ignore
 *
 * IDEAS for extensions:
 * - .script/.endscript pair
 * - .define/.enddefine pair to create long and multi-line variables
 * - Variables may have arguments (==> macros), e.g.
 *   twice(x) = $x$x
 * - Call macros: ${twice(argument)}
 * - .block/.endblock pair as a scope for local entities
 * - .local to declare local variables
 *   This will be very tricky. When the scope of a block is left, references
 *   to local variables remain intact. Every ${...} reference points to the
 *   local scope where it lexically occurs.
 * - Compatibility with GNU make
 *)


(* Optimized hash tables: *)
module S = struct
  type t = string * int  
	     (* The string and its hash value *)
  let equal ((a,ah):t) ((b,bh):t) =
    ah == bh && a = b
  let hash (x,xh) =
    xh
  let compare ((a,ah):t) ((b,bh):t) =
    match ah-bh with
	0 -> Pervasives.compare a b
      | n -> n
  let mk (x:string) =
    (x, Hashtbl.hash x)
  let dest (x,xh) =
    x
end
;;


module H = Hashtbl.Make(S)
;;


module SSet = Set.Make(S)
;;

open Printf;;
open Godi_types;;


type location =
    { loc_filename : string;
      loc_linenumber : int;
    }

type make_cmd =
    M_assign of     (lhs * value)
  | M_assign_app of (lhs * value)
  | M_assign_opt of (lhs * value)
  | M_assign_exp of (lhs * value)
  | M_assign_sh  of (lhs * value)
  | M_push of S.t
  | M_pop
  | M_undef of S.t
  | M_if of (expr * loc_make_cmd list * loc_make_cmd list)
  | M_for of (S.t list * value * loc_make_cmd list)
  | M_include of (bool * value)
  | M_sys_include of (bool * value)
  | M_dep of (value * value)
  | M_forced_dep of (value * value)
  | M_concat_dep of (value * value)
  | M_shell of value

and loc_make_cmd = location * make_cmd

and lhs =
    LHS_val of value
  | LHS_hashed_var of S.t

and value =
  | V_lit of string
  | V_esc of char
  | V_var of varspec
  | V_hashed_var of hashed_varspec
  | V_replaced_var of replaced_varspec
  | V_list of value list

and expr =
    E_or of (expr list)
  | E_and of (expr list)
  | E_not of expr
  | E_defined of value
  | E_make of value
  | E_exists of value
  | E_empty of varspec
  | E_target of value
  | E_commands of value
  | E_num_eq of (value * value)
  | E_num_ne of (value * value)
  | E_num_le of (value * value)
  | E_num_lt of (value * value)
  | E_num_ge of (value * value)
  | E_num_gt of (value * value)
  | E_str_eq of (value * value)
  | E_str_ne of (value * value)

and varspec = value * modifier list

and replaced_varspec = string * string * modifier list

and hashed_varspec = S.t * modifier list

and modifier =
    MO_E
  | MO_H
  | MO_M of value
  | MO_N of value
  | MO_native
  | MO_O
  | MO_Q
  | MO_R
  | MO_tl
  | MO_tu
  | MO_S of (value * value * subst_modifier)
  | MO_C of (value * value * subst_modifier)
  | MO_T
  | MO_u
  | MO_if of (value * value)
  | MO_subst of (value * value)
  | MO_U of value
  | MO_D of value
  | MO_L
  | MO_P
  | MO_cmd of value
  | MO_sh
  | MO_map of (value * value)

and subst_modifier =
    { sm_first : bool;   (* Only first word is subject to substitution *)
      sm_global : bool;  (* All occurrences in the word/words are replaced *)
    }
;;


type magic_context =
    { mutable magic_target : string option;
        (* The target currently made *)
      mutable magic_target_suffix : string option;
        (* The suffix of [magic_target], as constructed by the implicit rule *)
      mutable magic_sources : string list;
        (* The current sources *)
      mutable magic_sources_ood : string list;
        (* The current sources that are out of date *)
      mutable magic_exec_level : int option;
        (* The execution level *)
    }
;;

type variable =
    V_normal of value
  | V_env of value
  | V_override of value
  | V_undef
  | V_magic of (location -> magic_context -> value)
;;


type rule =
    { rule_target : string;
      rule_target_suffix : string;
      mutable rule_sources : string list;
      mutable rule_atts : string list;
      mutable rule_commands: (location * value) list;
      rule_type : [ `Dep | `Forced_dep | `Concat_dep ];
      mutable rule_file_type : [ `File | `Phony | `Join ];
      mutable rule_ignore_flag : bool;
      mutable rule_made_flag : bool;
      mutable rule_force_exec_flag : bool;
      mutable rule_optional_flag : bool;
      mutable rule_silent_flag : bool;
      mutable rule_system_flag : bool;
    }
;;


type command =
    { command : string;
        (* The command string, to be executed by sh *)
      env : (string * string) list;
        (* Additional environment variables *)
      directory : Godi_file.filename option;
        (* maybe change directory *)
    }
;;


type context =
    { vars : variable H.t;
        (* Current values of variables *)
      exported_vars : string list;
      rules : rule list H.t;
        (* The rules, hashed by their targets *)
      final_targets : string list;
        (* The targets to finally create (corresponds to the targets
	 * mentioned on the command line)
	 *)
      mutable default_target : string option;
	(* The default target that is made when there are no
	 * final_targets
	 *)
      mutable sys_mk_path : Godi_file.filename list;
        (* The path where Makefiles are searched included with <...> notation *)
      mutable user_mk_path : Godi_file.filename list;
        (* The path where Makefiles are searched included with "..." notation *)
      mutable curdir : string;
        (* The directory where [make] is invoked *)
      mutable objdir : Godi_file.filename;
        (* The directory serving as virtual base directory *)
      capture_sh_cmd : context -> command -> (string * Godi_exec.process_status);
      call_sh_cmd : context -> command -> Unix.process_status;

      mutable global_ignore_flag : bool;   (* .IGNORE for all rules *)
      mutable global_silent_flag : bool;   (* .SILENT for all rules *)
      mutable global_precious_flag : bool; (* .PRECIOUS for all rules *)

      mutable suffixes : string list;      (* list of suffixes *)

      mutable precious_targets : string list;

      console : console;
        (* Target for warnings etc. *)

      trace : bool;
      trace_script : bool;

      warn_on_env_access : bool;
        (* Whether to print a warning when an environment variable is
	 * accessed that is not listed in GODI_MAKE_ENV.
	 *)
    }
;;


type parsed_cmd =
    (* Used for [End_of_section], see below *)
    C_undef of string
  | C_if    of expr
  | C_else
  | C_elif  of expr
  | C_endif
  | C_for   of (string list * value)
  | C_endfor
  | C_include of (bool * value)
  | C_sys_include of (bool * value)
;;

(**********************************************************************)
(* Exceptions                                                         *)
(**********************************************************************)

exception Parse_error of (location * string);;
  (* exported *)

exception Eval_error of (location * string);;
  (* exported *)

exception Exec_error of (location * string);;
  (* exported *)

exception File_not_found of string;;
  (* exported *)

exception Target_not_found of string ;;
  (* exported *)

exception End_of_section of (location * parsed_cmd);;
  (* Internally used to exit nested parsing when a conditional section ends
   * (e.g. when .else is found). The parsed_cmd is the command functioning
   * as closing brace.
   *)

exception Bad_modifier;;
  (* Internally used while parsing modifiers to indicate that the function
   * could not parse the modifier and that another way of parsing should
   * be tried next
   *)

exception Unknown_directive;;
  (* Internally used to signal that the current line is not a .directive *)

exception Out_of_date
  (* exported *)


let catch_user_error ?(pref_user_error=false) f arg =
  let location loc =
    "File " ^ loc.loc_filename ^ ", line " ^
    string_of_int loc.loc_linenumber
  in
  try
    f arg
  with
      Parse_error(loc, msg) when pref_user_error ->
	user_error(sprintf "Parse error: %s: %s" (location loc) msg)
    | Eval_error(loc, msg) when pref_user_error ->
	user_error(sprintf "Eval error: %s: %s" (location loc) msg)
    | File_not_found msg when pref_user_error ->
	user_error(sprintf "File not found: %s" msg)
    | Sys_error msg when pref_user_error ->
	user_error(sprintf "System error: %s" msg)
    | Exec_error(loc,msg) when pref_user_error ->
	user_error(sprintf "Exec error: %s: %s" (location loc) msg)
    | Target_not_found msg when pref_user_error ->
	user_error(sprintf "Target not found: %s" msg)
;;

(**********************************************************************)
(* Regular expressions                                                *)
(**********************************************************************)

let operator_re =
  Str.regexp "[+?:!]?=\\|::?\\|!";;

let comment_re =
  (* Str.regexp "[ \t]+#";; *)
  Str.regexp "#";;

let old_include_re =
  Str.regexp "^[ \t]*\\(-?\\)include[ \t]+\\([^ \t]+\\)[ \t]*$";;

let empty_re =
  Str.regexp "^[ \t\r\n]*$"

let non_ws_re =
  Str.regexp "[^ \t]";;

let dollar_re =
  Str.regexp "[$]";;

let dot_cmd_re =
  Str.regexp "^\\.[ \t]*\\([a-z-]+\\)\\([ \t]+\\|$\\)";;

let word_re =
  Str.regexp "[^ \t\r\n]+";;

let alphanum_word_re =
  (* Include only chars that do not cause ambiguity in expressions *)
  Str.regexp "[^ \t()|&!$\"{}]*";;

let arg_re =
  Str.regexp "([ \t]*\\([^ \t)]+\\)[ \t]*)";;

let ws_re =
  Str.regexp "[ \t\r\n]+";;

let or_op_re =
  Str.regexp "[|][|]";;

let and_op_re =
  Str.regexp "&&";;

let cmp_op_re =
  Str.regexp "==\\|!=\\|<=\\|<\\|>=\\|>" ;;

let dec_number_re =
  Str.regexp "[-+]?[0-9]*\\(\\.[0-9]+\\|[0-9]\\)\\([Ee][-+]?[0-9]+\\)?";;

let hex_number_re =
  Str.regexp "0[xX][0-9a-fA-F]+";;

let dquote_re =
  Str.regexp "\"";;

let sh_newline_re =
  Str.regexp "\n";;

let sh_trailing_newline_re =
  Str.regexp "\n$";;

let env_re =
  Str.regexp "^\\([^=]+\\)=\\(.*\\)$";;

let special_source_re = 
  Str.regexp "\\.[A-Z0-9_]+$";;


(**********************************************************************)
(* General auxiliary functions                                        *)
(**********************************************************************)

let extension_re = Str.regexp ".*\\(\\.[^./]*\\)$"

let default_target_suffix t =
  if Str.string_match extension_re t 0 then
    Str.matched_group 1 t
  else
    ""
;;


let special_cmds =
  (* Output of these commands is cached because it is assumed that it
   * never changes.
   *)
  [ "/bin/uname -m", ref None;
    "/bin/uname -n", ref None;
    "/bin/uname -r", ref None;
    "/bin/uname -s", ref None;
    "/bin/uname -p", ref None;
    "/usr/bin/uname -m", ref None;
    "/usr/bin/uname -n", ref None;
    "/usr/bin/uname -r", ref None;
    "/usr/bin/uname -s", ref None;
    "/usr/bin/uname -p", ref None;
  ]
;;


let really_run_sh_cmd capture_flag ctx cmd =
  (* TODO: Create environment, especially set PWD to objdir *)
  let buf = Buffer.create 500 in
  try
    let stdout =
      if capture_flag then
	Some(`Buffer buf)
      else
	None in   (* CHECK: maybe we need more flexible options for stdout *)
    Godi_exec.call_cmd
      ~env:cmd.env
      ?stdout
      ~stderr:(ctx.console.protocol :> Godi_exec.stderr_collector)
      ?cd:cmd.directory
      (Godi_exec.shell ~setx:ctx.trace_script cmd.command);
    (Buffer.contents buf, Godi_exec.WEXITED 0)
  with
      Godi_exec.Command_failure(_,_,code) ->
	("", code)
;;


let capture_sh_cmd ctx cmd =
  (* TODO: Do this optimization less hack-ish *)
  try
    let r = List.assoc cmd.command special_cmds in
    match !r with
	Some s -> s
      | None   ->
	  let s = really_run_sh_cmd true ctx cmd in
	  r := Some s;
	  s
  with
      Not_found ->
	really_run_sh_cmd true ctx cmd
;;


let call_sh_cmd ctx cmd =
  let (_, code) = really_run_sh_cmd false ctx cmd in
  code
;;


let check_pstatus loc cmd p =
  (* Raise Eval_error when p is a non-zero exit status *)
  match p with
      Godi_exec.WEXITED 0 ->
	()
    | Godi_exec.WEXITED n ->
	raise(Eval_error(loc, ("Command returned with non-zero exit code: " ^ cmd)))
    | Godi_exec.WSIGNALED n ->
	raise(Eval_error(loc, ("Command terminated by signal: " ^ cmd)))
    | Godi_exec.WSTOPPED _ ->
        assert false
;;


let split_in_words s =
  (* Split [s] into whitespace-separated words *)
  Str.split ws_re s
;;



let generate_files ?base_dir loc arg =
  (* [arg] is a space-separated list of glob expressions. Returns the
   * expanded expressions. The meta characters *, ?, [ are expanded
   * when they actually match file names, otherwise the sub expression
   * becomes empty
   *)
  let base_dir' =
    match base_dir with
      | None -> None
      | Some d -> Some (Godi_file.slashname d) in
  let words = split_in_words arg in
  let gexprs =
    List.map
      (fun s ->
	 try (s, Godi_match.parse_glob_expr s)
	 with
	     Godi_match.Bad_glob_expr _ ->
	       raise(Eval_error(loc, ("Bad glob expression: " ^ s)))
	   | Godi_match.Unsupported_expr _ ->
	       raise(Eval_error(loc, ("Unsupported glob expression: " ^ s)))
      )
      words in
  let files =
    List.flatten
      (List.map
	 (fun (s, gexpr) ->
	    try
	      Godi_match.glob ?base_dir:base_dir' ~mode:`All_paths gexpr
	    with
	      | Sys_error msg ->
		  raise(Sys_error (msg ^ " (glob)"))
	 )
	 gexprs
      ) in
  
  files
;;


let split_regularly n l =
  (* Divide l into sublists of length n *)
  let rec collect k l =
    if k = 0 then
      let g = collect n l in
      [] :: g
    else
      match l with
	  x :: l' ->
	    let g = collect (k-1) l' in
	    ( match g with
		  y :: g' ->
		    (x::y) :: g'
		| _ ->
		    assert false
	    )
	| [] ->
	    if k = n then
	      []
	    else
	      assert false
  in
  collect n l
;;


let rec preceding_backslashes k s =
  (* Returns the number of backslashes before s.[k] *)
  if k > 0 then
    if s.[k-1] = '\\' then
      1 + preceding_backslashes (k-1) s
    else
      0
  else
    0
;;


let strip ?(left=false) ?(right=false) ?(right_bs=false) s =
  (* ~left: strip whitespace at the beginning
   * ~right: strip whitespace at the end
   * ~right_bs: backslashes are honoured at the right end
   *)
  let n = String.length s in

  let i =
    if left then
      try Str.search_forward non_ws_re s 0 with Not_found -> n
    else
      0
  in

  let j =
    if right && n>0 then
      let j0 =
	try
	  Str.search_backward non_ws_re s (n-1) + 1
	with
	    Not_found -> 0
      in
      (* If s.[j0-1] is a backslash: Check how many backslashes we have, maybe
       * we have to include the following character
       *)
      if right_bs && j0 > 0 && s.[j0-1] = '\\' && j0 < n then (
	let m = preceding_backslashes j0 s in
	if m land 1 = 1 then
	  j0+1
	else
	  j0
      )
      else j0
    else
      n
  in

  if i > j then
    ""
  else
    if i = 0 && j = n then
      s                                (* avoid making a copy *)
    else
      String.sub s i (j-i)
;;


let remove_comment s =
  (* A '#' is only a comment character if the number of preceding
   * backslashes is zero or even
   *)
  let rec loop k =
    try
      let i = Str.search_forward comment_re s k in
      if (preceding_backslashes i s) land 1 = 0 then
	String.sub s 0 i
      else
	loop (k+1)
    with
	Not_found ->
	  s
  in
  loop 0
;;


let value_list v =
  (* Returns a value as a list of component values where
   * - no component value is V_list
   * - there are no adjacent V_lit
   * - there is no V_lit ""
   *)
  let rec accumulate acc vl =
    match vl with
	[] ->
	  acc
      | (V_list ivl)::vl' ->
	  accumulate acc (ivl @ vl')
      | (V_esc c)::vl' ->
	  accumulate ((V_esc c)::acc) vl'
      | (V_var vs)::vl' ->
	  accumulate ((V_var vs)::acc) vl'
      | (V_replaced_var vs)::vl' ->
	  accumulate ((V_replaced_var vs)::acc) vl'
      | (V_hashed_var vs)::vl' ->
	  accumulate ((V_hashed_var vs)::acc) vl'
      | (V_lit s)::vl' ->
	  if s = "" then
	    accumulate acc vl'
	  else
	    ( match acc with
		  (V_lit u)::acc' ->
		    accumulate ((V_lit (u^s))::acc') vl'
		| _ ->
		    accumulate ((V_lit s)::acc) vl'
	    )
  in
  match v with
      V_list vl -> List.rev (accumulate [] vl)
    | _         -> List.rev (accumulate [] [v])
;;


let split_value_by_operator v =
  (* Returns (lhs,op,rhs) if there is an assignment or dependency operator
   * op in v, or fails
   *)
  let rec find_operator acc vl =
    match vl with
	[] ->
	  failwith "Syntax error"
      | (V_lit s) :: vl' ->
	  ( try
	      let k = Str.search_forward operator_re s 0 in
	      let j = Str.match_end() in
	      let op = Str.matched_string s in
	      let s1 = strip ~right:true
			 (String.sub s 0 k) in
	      let s2 = strip ~left:true
			 (String.sub s j (String.length s - j)) in
	      let l1 = if s1 = "" then [] else [V_lit s1] in
	      let l2 = if s2 = "" then [] else [V_lit s2] in
	      (V_list(List.rev (l1 @ acc)), op, V_list(l2 @ vl'))
	    with
		Not_found ->
		  find_operator ((V_lit s)::acc) vl'
	  )
      | v :: vl' ->
	  find_operator (v :: acc) vl'
  in

  let vl = value_list v in
  find_operator [] vl
;;


let search_file basedir path name =
  (* If [name] does not begin with "/", it will be searched in [path].
   * Returns either the full path name, or raises File_not_found.
   * [basedir] is prepended to relative names in [path].
   *)
  try
    ( let parsed_name = Godi_file.filename name in
      match Godi_file.nameclass parsed_name with
	| `POSIX_absolute | `WIN32_drive | `WIN32_remote ->
	    if Godi_file.file_exists parsed_name then
	      parsed_name
	    else
	      raise(File_not_found name)
	| `Relative ->
	    ( try
		let prefix =
		  List.find (fun prefix ->
			       let abs_prefix = Godi_file.concat basedir prefix in
			       let full_path = Godi_file.addname abs_prefix name in
			       Godi_file.file_exists full_path) path in
		let abs_prefix = Godi_file.concat basedir prefix in
		Godi_file.addname abs_prefix name
	      with
		  Not_found ->
		    raise(File_not_found name)
	    )
	| `Error ->
	    raise(File_not_found name)
    )
  with
    | Sys_error msg ->
	raise(Sys_error(msg ^ " (search_file)"))
;;


let rec const_value v =
  (* Returns the constant string value of v, or raises Not_found if there
   * are variables
   *)
  match v with
      V_lit s  -> s
    | V_esc c  -> "\\" ^ String.make 1 c
    | V_list l -> String.concat "" (List.map const_value l)
    | V_var(_,_) -> raise Not_found
    | V_hashed_var(_,_) -> raise Not_found
    | V_replaced_var(_,s,[]) -> s
    | V_replaced_var(_,s,_) -> raise Not_found
;;


let create_var (v,mods) =
  (* Creates either a V_var or a V_hashed_var variable reference *)
  try
    let s = const_value v in
    V_hashed_var(S.mk s, mods)
  with
      Not_found -> V_var (v,mods)
;;


(**********************************************************************)
(* The parser                                                         *)
(**********************************************************************)


let loc_unknown = { loc_filename = ""; loc_linenumber = 0 } ;;

let parse_error msg =
  raise(Parse_error(loc_unknown, msg))
;;


(* Character categories:
 * 'D': delimiters
 * 'X': inner dollar expansion
 * 'Y': inner dollar expansion, ignore illegal '$' notation (for C and
 *      S modifiers)
 * 'E': character may be escaped
 * ' ': normal character
 *)

let rec parse_until ?(keep_escaped = false) ?(need_delim = true)
                    ?(recognize_backslash = true)
                    s charcat k =
  (* Parses sequence until one of the delims mentioned in charcat is
   * found (~need_delim), or until the end of the string is reached
   * (not ~need_delim). Parsing starts at position k of string s.
   *
   * Backslashes are always considered as escape characters.
   * (See note at end of .ml file.) By default, the allowed escape
   * sequences are resolved and returned as normal V_lit values. If
   * ~keep_escaped, the backslashed character is copied to the
   * output list as V_esc value.
   *
   * Escape sequences: Backslash plus:
   *  - '#'
   *  - Delimiter
   *  - Characters of category 'E'
   *
   * Unless not ~need_delim, it is an error if the end of the string
   * is found.
   *
   * The parsed sequence may consist of characters and dollar
   * variables, e.g. "<text>${variable}<text>". The sequence
   * is returned as pair (vlist,j), where vlist is the list of
   * values, and j is the position of the delimiter.
   *
   * ~_start: internally used: denotes the beginning of the current
   * character sequence.
   *)

  let n = String.length s in

  let rec parse_loop start k =
    let v0 =
      if k = start then
	[]
      else
	[V_lit (String.sub s start (k - start))] in
    if k < n then (
      let c = s.[k] in
      let cat = charcat.[ Char.code c ] in
      if cat = 'D' then (  (* delimiter found *)
	(v0,k)    (* stop here! *)
      )
      else if cat = 'X' then  (* inner dollar expansion found *)
	let (v,k') = parse_dollar s k in
	let (l,k'') = parse_loop k' k' in
	(v0 @ (v :: l), k'')
      else if cat = 'Y' then  (* inner dollar expansion found *)
	let (v,k') = parse_dollar ~tolerant:true s k in
	let (l,k'') = parse_loop k' k' in
	(v0 @ (v :: l), k'')
      else if c = '\\' && recognize_backslash then (
	if k >= n-1 then
	  parse_error ("Bad backslash");
	(* Should not happen, because backslashes at line ends are
	 * already handled.
	 *)
	let c' = s.[k+1] in
	let cat' = charcat.[ Char.code c' ] in
	let v2 =
	  match cat' with
	      ' ' | 'X' | 'Y' ->
		if keep_escaped then
		  V_esc c'
		else
		  if c' = '#' then
		    (* The only character that is automatically resolved *)
		    V_lit (String.make 1 c')
		  else
		    (* These cannot be resolved. *)
		    V_lit ("\\" ^ String.make 1 c')
	    |	'E' | 'D' ->
		  if keep_escaped then
		    V_esc c'
		  else
		    V_lit (String.make 1 c')
	    | _ ->
		assert false in
	let (l, k') = parse_loop (k+2) (k+2) in
	(v0 @ (v2 :: l), k')
      )
      else
	parse_loop start (k+1)
    )
    else
      if need_delim then
	parse_error ("Delimiter not found")
      else
	(v0,k)
  in
  parse_loop k k

and parse_varspec s k delim =
  (* Parses a varspec at s.[k] until delim is found. Returns
   * (vspec,j) where vspec is the parsed varspec, and j is the position
   * of delim.
   *
   * A varspec is the construct ${name:modifiers} or $(name:modifiers).
   * k denotes the position of the first character of the name.
   * delim is either '}' or ')'. - This function cannot parse one-letter
   * varspecs like $x.
   *)

  let n = String.length s in

  let parse_name k =
    (* Parse ${name:mods} or $(name:mods) until the colon or the
     * closing parenthesis is found.
     * k is the position of the first character of the name.
     * Returns (vlist,j) (like parse_until)
     *)
    let charcat = String.make 256 ' ' in
    charcat.[ Char.code ':' ] <- 'D';    (* Delimiter *)
    charcat.[ Char.code '$' ] <- 'X';    (* Enable inner expansion *)
    charcat.[ Char.code '\\' ] <- 'E';   (* Interpret double-backslash *)
    charcat.[ Char.code delim ] <- 'D';  (* Delimiter *)
    let l, j = parse_until s charcat k in
    (V_list l, j)
  in

  let parse_pattern k =
    (* Parses a pattern p after :M or :N until the delimiter is found.
     * Returns (p, j) where j is the position of the delimiter.
     *)
    let charcat = String.make 256 ' ' in
    (* CHECK: Is '$' special? *)
    charcat.[ Char.code ':' ] <- 'D';
    charcat.[ Char.code delim ] <- 'D';
    let l, j = parse_until ~keep_escaped:true s charcat k in
    (V_list l, j)
  in

  let parse_value k =
    (* For U, D *)
    let charcat = String.make 256 ' ' in
    charcat.[ Char.code ':' ] <- 'D';
    charcat.[ Char.code '$' ] <- 'X';
    charcat.[ Char.code '\\' ] <- 'E';
    charcat.[ Char.code delim ] <- 'D';
    let (l, j) = parse_until s charcat k in
    (V_list l, j)
  in

  let parse_cmd k =
    (* For ! *)
    let charcat = String.make 256 ' ' in
    charcat.[ Char.code '!' ] <- 'D';
    charcat.[ Char.code '$' ] <- 'X';
    let (v,j) = parse_until s charcat k in
    if j+1 >= n then raise Bad_modifier;
    if s.[j+1] <> ':' && s.[j+1] <> delim then raise Bad_modifier;
    (V_list v, j+1)
  in

  let parse_if k =
    (* For ? : *)
    let charcat = String.make 256 ' ' in
    charcat.[ Char.code '$' ] <- 'X';
    charcat.[ Char.code ':' ] <- 'D';
    charcat.[ Char.code '\\' ] <- 'E';
    let (v1,j1) = parse_until s charcat k in
    if j1+1 >= n then raise Bad_modifier;
    charcat.[ Char.code delim ] <- 'D';
    let (v2,j2) = parse_until s charcat (j1+1) in
    (V_list v1, V_list v2, j2)
  in

  let parse_map k =
    (* For @ *)
    let charcat = String.make 256 ' ' in
    charcat.[ Char.code '@' ] <- 'D';
    charcat.[ Char.code '$' ] <- 'X';
    charcat.[ Char.code '\\' ] <- 'E';
    let (v1,j1) = parse_until s charcat k in
    if j1+1 >= n then raise Bad_modifier;
    let (v2,j2) = parse_until s charcat (j1+1) in
    (V_list v1, V_list v2, j2+1)
  in

  let rec parse_subst_modifier k =
    (* Parse the "1" and "g" flags at the end of C, S *)
    let null_sm = { sm_first = false; sm_global = false } in
    if k < n then
      match s.[k] with
	  '1' -> let (sm, k') = parse_subst_modifier (k+1) in
                 ( { sm with sm_first = true }, k' )
	| 'g' ->let (sm, k') = parse_subst_modifier (k+1) in
                 ( { sm with sm_global = true }, k' )
	| _   -> (null_sm, k)
    else
      (null_sm, k)
  in

  let parse_repl k =
    (* For C, S *)
    if k >= n then raise Bad_modifier;
    let d = s.[k] in
    let charcat = String.make 256 ' ' in
    charcat.[ Char.code '$' ] <- 'Y';
    charcat.[ Char.code d ] <- 'D';
    let (v1,j1) = parse_until ~keep_escaped:true s charcat (k+1) in
    if j1+1 >= n then raise Bad_modifier;
    let (v2,j2) = parse_until ~keep_escaped:true s charcat (j1+1) in
    (* if j2+1 >= n then raise Bad_modifier; ??? *)
    let (sm, j3) = parse_subst_modifier (j2+1) in
    (V_list v1, V_list v2, sm, j3)
  in

  let try_also_subst f_parse k =
    (* Parse the modifier with f_parse, but try also MO_subst if f_parse
     * fails.
     * CHECK: Compare this solution with what "bmake" does
     *)
    try
      f_parse k
    with
	Bad_modifier ->
	  if s.[k] = ':' then (
	    (* Try also a oldstr=newstr modifier. *)
	    let charcat = String.make 256 ' ' in
	    charcat.[ Char.code delim ] <- 'D';
	    charcat.[ Char.code '=' ] <- 'D';
	    let (v_oldstr, k') = parse_until s charcat (k+1) in
	    if s.[k'] <> '=' then raise Bad_modifier;
	    charcat.[ Char.code '=' ] <- ' ';
	    let (v_newstr, k'') = parse_until s charcat (k'+1) in
	    ( [ MO_subst(V_list v_oldstr, V_list v_newstr) ], k'' )
	  )
	  else
	    parse_error "Bad variable modifier"
	      (* We do not propagate the Bad_modifier exception up.
	       * This would be a bit surprising...
	       *)
  in

  let rec parse_modifiers k =
    (* raises Bad_modifier if the modifier at k could not be parsed,
     * and it might be possible that this is a MO_subst modifier.
     *)
    if s.[k] = ':' then (
      if k+1 >= n then raise Bad_modifier;
      let m, k' =
	match s.[k+1] with
	    'E' -> (MO_E, k+2)
	  | 'H' -> (MO_H, k+2)
	  | 'M' -> let pat, j = parse_pattern (k+2) in
	           (MO_M pat, j)
	  | 'N' -> let pat, j = parse_pattern (k+2) in
	           (MO_N pat, j)
	  | 'n' -> (MO_native, k+2)
	  | 'O' -> (MO_O, k+2)
	  | 'Q' -> (MO_Q, k+2)
	  | 'R' -> (MO_R, k+2)
	  | 't' ->
	      if k+2 >= n then raise Bad_modifier;
	      let c = s.[k+2] in
	      ( match c with
		    'l' -> (MO_tl, k+3)
		  | 'u' -> (MO_tu, k+3)
		  | _   -> raise Bad_modifier
	      )
	  | 'S' -> let (v1,v2,sm,j) = parse_repl (k+2) in
	           (MO_S(v1,v2,sm), j)
	  | 'C' -> let (v1,v2,sm,j) = parse_repl (k+2) in
	           (MO_C(v1,v2,sm), j)
	  | 'T' -> (MO_T, k+2)
	  | 'u' -> (MO_u, k+2)
	  | '?' -> let v1, v2, j = parse_if (k+2) in
	           (MO_if(v1,v2), j)
	  | '@' -> let v1, v2, j = parse_map (k+2) in
		     (MO_map(v1,v2), j)
	  | 'U' -> let v, j = parse_value (k+2) in
	           (MO_U v, j)
	  | 'D' -> let v, j = parse_value (k+2) in
	           (MO_D v, j)
	  | 'L' -> (MO_L, k+2)
	  | 'P' -> (MO_P, k+2)
	  | '!' -> let v, j = parse_cmd (k+2) in
	           (MO_cmd v, j)
	  | 's' ->
	      if k+2 >= n then raise Bad_modifier;
	      let c = s.[k+2] in
	      ( match c with
		    'h' -> (MO_sh, k+3)
		  | _   -> raise Bad_modifier
	      )
	  | ':' ->
	      parse_error "Limitation: Assignment not supported here"
	  | _ ->
	      raise Bad_modifier
      in
      let (mods', k'') = try_also_subst parse_modifiers k' in
      (m :: mods', k'')
    )
    else if s.[k] = delim then
      ( [], k )
    else
      raise Bad_modifier
  in

  (* cont parse_varspec: *)

  try
    let (v_name, k1) = parse_name k in
    let (mods, k2) = try_also_subst parse_modifiers k1 in
    ( (v_name, mods), k2 )
  with
      Bad_modifier ->
	parse_error "Bad variable modifier"

(* end parse_varspec *)

(* --- *)

and parse_dollar ?(tolerant = false) s k =
  (* s.[k] is a dollar character. Returns (val,j) where j is the
   * position after the dollar notation.
   *
   * Precondition: s.[k] = '$'
   *
   * ~tolerant: If true, illegal dollar notations are accepted, and returned
   * as normal literal values
   *)
  assert(s.[k] = '$');
  if k >= String.length s - 1 then (
    if not tolerant then
      parse_error "Found '$' at end of line";
    (V_lit "$", k+1)
  )
  else
    match s.[k+1] with
	'$' ->
	  (V_lit "$", k+2)
      | '{' ->
	  let v, j = parse_varspec s (k+2) '}' in
	  let v' = create_var v in
	  (v', j+1)
      | '(' ->
	  let v, j = parse_varspec s (k+2) ')' in
	  let v' = create_var v in
	  (v', j+1)
      | 'A'..'Z' | 'a'..'z' | '>' | '<' | '?' | '*' | '@' ->
	  (V_hashed_var(S.mk(String.make 1 s.[k+1]),[]), k+2)
      | _ ->
	  if not tolerant then
	    parse_error "Illegal '$' notation";
	  (V_lit "$", k+1)
;;


let parse_value ?(pos=0) (* ?len *) s =
  let charcat = String.make 256 ' ' in
  charcat.[ Char.code '$' ] <- 'X';
  let (l, _) = parse_until ~need_delim:false s charcat pos in
  V_list l
;;

(* --- *)

let parse_delimited_string s k delim =
  let charcat = String.make 256 ' ' in
  charcat.[ Char.code delim ] <- 'D';
  let (l, j) = parse_until s charcat k in
  let u =
    String.concat
      ""
      (List.map
	 (function
	      V_lit s -> s
	    | V_var _ -> assert false
	    | V_hashed_var _ -> assert false
	    | V_esc _ -> assert false
	    | V_list _ -> assert false  (* hope so *)
	    | V_replaced_var _ -> assert false
	 )
	 l) in
  (u, j+1)
;;


let parse_string_literal s k =
  if s.[k] <> '"' then parse_error "Syntax error";
  parse_delimited_string s (k+1) '"'
;;


let parse_include_literal s k =
  let charcat = String.make 256 ' ' in
  charcat.[ Char.code '$' ] <- 'X';
  if s.[k] = '"' then (
    charcat.[ Char.code '"' ] <- 'D';
    let (l, j) = parse_until s charcat (k+1) in
    ((V_list l, (j+1)), false)
  ) else
    if s.[k] = '<' then (
      charcat.[ Char.code '>' ] <- 'D';
      let (l, j) = parse_until s charcat (k+1) in
      ((V_list l, (j+1)), true)
    )
    else
      parse_error "File to include must be surrounded by \"...\" or <...>"
;;


let skip_ws s k =
  if Str.string_match ws_re s k then
    Str.match_end()
  else
    k
;;


let parse_word s k =
  (* Returns (word, k'), where k' is the position after the word (and after
   * whitespace)
   *)
  if Str.string_match word_re s k then
    let u = Str.matched_string s in
    (u, skip_ws s (Str.match_end()))
  else
    parse_error "Syntax error"
;;


let parse_alphanum_word s k =
  (* Returns (word, k'), where k' is the position after the word (and after
   * whitespace)
   *)
  if Str.string_match alphanum_word_re s k then
    let u = Str.matched_string s in
    (u, skip_ws s (Str.match_end()))
  else
    parse_error "Syntax error"
;;


let parse_argument s k =
  if k >= String.length s || s.[k] <> '(' then
    parse_error "Syntax error";
  let charcat = String.make 256 ' ' in
  charcat.[ Char.code ')' ] <- 'D';
  charcat.[ Char.code '$' ] <- 'X';
  let (l, j) = parse_until s charcat (k+1) in
  (V_list l, skip_ws s (j+1))
;;


let rec parse_or_expr p s k =
  (* Returns (expr, k'), where k' is the position after the || expression *)
  (* p: A function that substitutes symbols *)
  let (e1, k') = parse_and_expr p s k in
  if Str.string_match or_op_re s k' then
    let (e2, k'') = parse_or_expr p s (k'+2) in
    match e2 with
	E_or l ->
	  (E_or (e1::l), k'')
      | _ ->
	  (E_or [e1;e2], k'')
  else
    (e1, k')

and parse_and_expr p s k =
  (* Returns (expr, k'), where k' is the position after the && expression *)
  let (e1, k') = parse_atom p s k in
  if Str.string_match and_op_re s k' then
    let (e2, k'') = parse_and_expr p s (k'+2) in
    match e2 with
	E_and l ->
	  (E_and (e1::l), k'')
      | _ ->
	  (E_and [e1;e2], k'')
  else
    (e1, k')

and parse_atom p s k =
  let k1 = skip_ws s k in
  if k1 >= String.length s then
    parse_error "Syntax error in expression";
  match s.[k1] with
      '!' ->
	let (e', k') = parse_atom p s (k1+1) in
	(E_not e', k')
    | '(' ->
	let (e', k') = parse_or_expr p s (k1+1) in
	if k' >= String.length s || s.[k'] <> ')' then
	  parse_error "Syntax error in expression: Missing ')'";
	let k'' = skip_ws s (k'+1) in
	(e', k'')
    | '$' ->
	let (var, k') = parse_dollar s k1 in
	let k'' = skip_ws s k' in
	if not (Str.string_match cmp_op_re s k'') then
	  parse_error "Syntax error in expression: comparison op expected";
	let op = Str.matched_string s in
	let j = skip_ws s (Str.match_end()) in
	if Str.string_match hex_number_re s j ||
	  Str.string_match dec_number_re s j then (
	    let v = V_lit (Str.matched_string s) in
	    let e =
	      match op with
		  "==" -> E_num_eq(var,v)
		| "!=" -> E_num_ne(var,v)
		| "<=" -> E_num_le(var,v)
		| "<"  -> E_num_lt(var,v)
		| ">=" -> E_num_ge(var,v)
		| ">"  -> E_num_gt(var,v)
		| _    -> assert false in
	    let j' = skip_ws s (Str.match_end()) in
	    (e,j')
	  )
	else (
	  if Str.string_match dquote_re s j then
	    let v, j' = parse_string_literal s j in
	    let e =
	      match op with
		  "==" -> E_str_eq(var,V_lit v)
		| "!=" -> E_str_ne(var,V_lit v)
		| "<=" -> parse_error "Op <= not supported for strings"
		| "<"  -> parse_error "Op < not supported for strings"
		| ">=" -> parse_error "Op >= not supported for strings"
		| ">"  -> parse_error "Op > not supported for strings"
		| _    -> assert false in
	    let j'' = skip_ws s j' in
	    (e,j'')
	  else
	    if j < String.length s && s.[j] = '$' then
	      let (var', j') = parse_dollar s j in
	      let e =
		match op with
		    "==" -> E_str_eq(var,var')
		  | "!=" -> E_str_ne(var,var')
		  | "<=" -> E_num_le(var,var')
		  | "<"  -> E_num_lt(var,var')
		  | ">=" -> E_num_ge(var,var')
		  | ">"  -> E_num_gt(var,var')
		  | _    -> assert false in
	      let j'' = skip_ws s j' in
	      (e,j'')
	    else
	      parse_error "Syntax error in expression: literal or variable expected";
	)
    | _ ->
	let (w, k') = parse_alphanum_word s k1 in
	( match w with
	      "defined" ->
		let (arg,k'') = parse_argument s k' in
		(E_defined arg, k'')
	    | "make" ->
		let (arg,k'') = parse_argument s k' in
		(E_make arg, k'')
	    | "exists" ->
		let (arg,k'') = parse_argument s k' in
		(E_exists arg, k'')
	    | "target" ->
		let (arg,k'') = parse_argument s k' in
		(E_target arg, k'')
	    | "commands" ->
		let (arg,k'') = parse_argument s k' in
		(E_commands arg, k'')
	    | "empty" ->
		if k' >= String.length s || s.[k'] <> '(' then
		  parse_error "Syntax error";
		let vspec, j = parse_varspec s (k'+1) ')' in
		let j' = skip_ws s (j+1) in
		(E_empty(vspec), j')
	    | _ ->
		(p w, k')
	)
;;


let parse_fd filename fd =
  (* Parses from in_channel [fd]. The [filename] is used for error messages
     only
   *)

  let line_number = ref 0 in         (* number of the line being parsed *)
  let line_number_counter = ref 1 in (* number of the next line *)

  let parse_error_at msg =
    let loc = { loc_filename = filename;
		loc_linenumber = !line_number } in
    raise(Parse_error(loc, msg))
  in

  let map_error f arg =
    try f arg
    with
	Parse_error(loc, msg) ->
	  let loc' =
	    { loc_filename = if loc.loc_filename = "" then filename
  	                                              else loc.loc_filename;
	      loc_linenumber = if loc.loc_linenumber = 0 then !line_number
  	                                                 else loc.loc_linenumber;
	    } in
	  raise(Parse_error(loc',msg))
  in


  (* --- *)

  let rec read_cont_line() =
    (* Read a line with continuation. May raise End_of_file. *)
    let line = 
      try
	input_line fd
      with
	| Sys_error msg -> raise(Sys_error(msg  ^ " (input_line)")) in
    incr line_number_counter;
    if line <> "" && line.[ String.length line - 1] = '\\' then (
      try
	let follow_line = strip ~left:true (read_cont_line()) in
	String.sub line 0 (String.length line - 1) ^ " " ^ follow_line
	  (* NOTE: In bmake, dependency lines can be continued with ';'.
	   * This is undocumented, and seems to be a compatibility feature.
	   * We omit it here.
	   *)
	  (* TODO: slow for longer continuations *)
      with
	  End_of_file ->
	    parse_error_at "Missing continuation line"
    )
    else
      line
  in

  (* --- *)

  let rec parse_for_vars s k =
    let (w, j) = parse_word s k in
    if w = "in" then
      ([], j)
    else
      if k <= String.length s then
	let (wl, j') = parse_for_vars s j in
	( w::wl, j' )
      else
	parse_error_at "Syntax error in .for statement"
  in

  let pfail w =
    parse_error_at "Syntax error" in

  let pdef w =
    E_defined(V_lit w) in

  let pmake w =
    E_make(V_lit w) in

  let parse_dot_cmd line =
    if Str.string_match dot_cmd_re line 0 then
      let cmd_name = Str.matched_group 1 line in
      let arg_pos = Str.match_end() in
      match cmd_name with
	  "undef" ->
	    let (w, k) = parse_word line arg_pos in
	    if k < String.length line then
	      parse_error_at "Syntax error";
	    C_undef w
	| "if" ->
	    let (e, k) = parse_or_expr pfail line arg_pos in
	    if k < String.length line then
	      parse_error_at "Syntax error";
	    C_if e
	| "ifdef" ->
	    let (e, k) = parse_or_expr pdef line arg_pos in
	    if k < String.length line then
	      parse_error_at "Syntax error";
	    C_if e
	| "ifndef" ->
	    let (e, k) = parse_or_expr pdef line arg_pos in
	    if k < String.length line then
	      parse_error_at "Syntax error";
	    C_if (E_not e)
	| "ifmake" ->
	    let (e, k) = parse_or_expr pmake line arg_pos in
	    if k < String.length line then
	      parse_error_at "Syntax error";
	    C_if e
	| "ifnmake" ->
	    let (e, k) = parse_or_expr pmake line arg_pos in
	    if k < String.length line then
	      parse_error_at "Syntax error";
	    C_if (E_not e)
	| "else" ->
	    if arg_pos < String.length line then
	      parse_error_at "Syntax error";
	    C_else
	| "elif" ->
	    let (e, k) = parse_or_expr pfail line arg_pos in
	    if k < String.length line then
	      parse_error_at "Syntax error";
	    C_elif e
	| "elifdef" ->
	    let (e, k) = parse_or_expr pdef line arg_pos in
	    if k < String.length line then
	      parse_error_at "Syntax error";
	    C_elif e
	| "elifndef" ->
	    let (e, k) = parse_or_expr pdef line arg_pos in
	    if k < String.length line then
	      parse_error_at "Syntax error";
	    C_elif (E_not e)
	| "elifmake" ->
	    let (e, k) = parse_or_expr pmake line arg_pos in
	    if k < String.length line then
	      parse_error_at "Syntax error";
	    C_elif e
	| "elifnmake" ->
	    let (e, k) = parse_or_expr pmake line arg_pos in
	    if k < String.length line then
	      parse_error_at "Syntax error";
	    C_elif (E_not e)
	| "endif" ->
	    C_endif
	| "for" ->
	    let (vars, k) = parse_for_vars line arg_pos in
	    let e = parse_value ~pos:k line in
	    C_for(vars, e)
	| "endfor" ->
	    C_endfor
	| "include"
	| "-include"
	| "sinclude" ->
	    if arg_pos >= String.length line then
	      parse_error_at "Argument is missing";
	    let ((s, k), is_sys) = parse_include_literal line arg_pos in
	    let ignore_errors = cmd_name <> "include" in
	    if is_sys then
	      C_sys_include(ignore_errors, s)
	    else
	      C_include(ignore_errors, s)
	| _ ->
	    raise Unknown_directive
    else
      raise Unknown_directive
  in

  (* --- *)

  let rec parse_dot_cmd_section (loc,cmd) =
    match cmd with
	C_undef v ->
	  (loc, M_undef (S.mk v))
      | C_if e ->
	  let (cmd'_opt, mlist) = parse_section() in
	  ( match cmd'_opt with
		None -> parse_error_at "Missing .endif"
	      | Some (loc', C_endif) ->
		  (loc, M_if(e, mlist, []))
	      | Some (loc', C_else) ->
		  let (cmd''_opt, mlist') = parse_section() in
		  ( match cmd''_opt with
			None ->
			  parse_error_at "Missing .endif"
		      | Some (loc'', C_endif) ->
			  (loc, M_if(e, mlist, mlist'))
		      | Some (_, C_elif _) ->
			  parse_error_at ".if must be ended by .endif, not .elif"
		      | Some (_, C_endfor) ->
			  parse_error_at ".if must be ended by .endif, not .endfor"
		      | _ ->
			  assert false
		  )
	      | Some (loc', C_elif e') ->
		  let m = parse_dot_cmd_section (loc', (C_if e')) in
		  (loc, M_if(e, mlist, [m]))
	      | Some (_, C_endfor) ->
		  parse_error_at ".if must be ended by .endif, not .endfor"
	      | _ ->
		  assert false
	  )
      | C_for(varlist, value) ->
	  let (cmd'_opt, mlist) = parse_section() in
	  ( match cmd'_opt with
		None -> parse_error_at "Missing .endfor"
	      | Some (loc', C_endfor) ->
		  let varlist' = List.map S.mk varlist in
		  (loc, M_for(varlist', value, mlist))
	      | Some (_, C_endif) ->
		  parse_error_at ".for must be ended by .endfor, not .endif"
	      | Some (_, C_else) ->
		  parse_error_at ".for must be ended by .endfor, not .else"
	      | Some (_, C_elif _) ->
		  parse_error_at ".for must be ended by .endfor, not .elif"
	      | _ ->
		  assert false
	  )
      | C_else
      | C_elif _
      | C_endif
      | C_endfor ->
	  raise (End_of_section (loc, cmd))
      | C_include(ignore_errors, s) ->
	  (loc, M_include(ignore_errors, s))
      | C_sys_include(ignore_errors, s) ->
	  (loc, M_sys_include(ignore_errors, s))

  (* --- *)

  and parse_shell_cmd line =
    let charcat = String.make 256 ' ' in
    charcat.[ Char.code '$' ] <- 'X';
    let (l, _) =
      parse_until ~need_delim:false ~recognize_backslash:false line charcat 1
    in
    M_shell (V_list l)

  (* --- *)

  and parse_cmd() : (location * make_cmd) =
    line_number := !line_number_counter;
    let line1 = read_cont_line() in
    (* Types of commands:
     * - Comment, empty line: ignore
     * - VARNAME=VALUE and variants
     * - .command
     * - include without . (as exception)
     * - target: targets + rule
     *)
    if line1 = "" || line1.[0] = '#' then
      parse_cmd()  (* quickly skip over empty and whole comment lines *)
    else
      let loc = { loc_filename = filename; loc_linenumber = !line_number } in
      if line1.[0] = '\t' then
	(* shell command *)
	(loc, parse_shell_cmd line1)
      else
	let line1 = remove_comment line1 in
	try
	  if line1 = "" || Str.string_match empty_re line1 0 then
	    parse_cmd()
	  else if line1.[0] = '.' then
	    let cmd = parse_dot_cmd line1 in  (* or Unknown_directive *)
	    parse_dot_cmd_section (loc, cmd)
	  else if Str.string_match old_include_re line1 0 then
	    (* old-style "include" directive *)
	    let arg = Str.matched_group 2 line1 in
	    let ignore_flag = Str.matched_group 1 line1 = "-" in
	    let value = parse_value arg in
	    (loc, M_include(ignore_flag, value))
	  else
	    raise Unknown_directive
	with
	    Unknown_directive ->
	      (* Remaining possibilities:
	       * - variable assignment
	       * - dependency line
	       * To find it out, we parse the line as value, and search whether
	       * there is an assignment or dependency operator. Note that a
	       * simple regexp does not work, e.g. an assignment
	       * ${v:a=b}=c must not be parsed as assigning "${v:a" the value
	       * "v}=c", but as assigning "${v:a=b}" the value "c".
	       *)
	      let line2 = strip ~right:true ~right_bs:true line1 in
	      let value = parse_value line2 in
	      let (lhs0, operator, rhs) =
		try
		  split_value_by_operator value
		with
		    Failure msg -> parse_error_at msg
	      in
	      let lhs =
		try LHS_hashed_var(S.mk(const_value lhs0))
		with Not_found -> LHS_val lhs0 in
	      let cmd =
		match operator with
		    "="  -> M_assign(lhs,rhs)
		  | "?=" -> M_assign_opt(lhs,rhs)
		  | ":=" -> M_assign_exp(lhs,rhs)
		  | "+=" -> M_assign_app(lhs,rhs)
		  | "!=" -> M_assign_sh(lhs,rhs)
		  | ":"  -> M_dep(lhs0,rhs)
		  | "::" -> M_concat_dep(lhs0,rhs)
		  | "!"  -> M_forced_dep(lhs0,rhs)
		  | _ ->
		      (* prerr_endline operator; *)
		      assert false in
	      (loc, cmd)

  (* --- *)

  and parse_section() : ((location * parsed_cmd) option *
                         (location * make_cmd) list) =
    (* Parses until EOF, .else, .elif*, .endif, and returns this line *)
    try
      let (loc, cmd) =
	map_error parse_cmd() in  (* or End_of_file, End_of_section *)
      let (following_line, cmds) = parse_section() in
      (following_line, (loc,cmd) :: cmds)
    with
	End_of_file             -> (None, [])
      | End_of_section(loc,cmd) -> (Some(loc,cmd), [])
  in

  (* cont parse_fd: *)

  ( try
      let (following_line, cmds) = parse_section() in
      if following_line <> None then
	parse_error_at "No conditional section to end here";
      cmds
    with
      | Sys_error msg ->
	  raise(Sys_error (msg ^ " (at line " ^ string_of_int !line_number 
			     ^ ")"))
      | e ->
	  prerr_endline("Exception at line " ^ string_of_int !line_number ^ 
			  ": "  ^ Printexc.to_string e);
	  raise e
  )
;;



let getcwd() =
  match Godi_file.getcwd() with
    | None ->
	raise(Sys_error "Cannot determine path of current working directory")
    | Some cwd ->
	cwd



let parse ?(pref_user_error=false) ?(system_scope=false) =
  (* prerr_endline ("Parsing " ^ filename); *)
  catch_user_error 
    ~pref_user_error 
    (fun filename ->
       let e_filename = Godi_file.slashname filename in
       if not (Godi_file.file_exists filename) then
	 raise(File_not_found e_filename);
       let fd = Godi_file.open_in `Text filename in
       let dirname_opt = Godi_file.dirname filename in
       let basename = Godi_file.basename filename in
       let parsedir =
	 match dirname_opt with
	   | None ->
	       (* We cannot parse directories... *)
	       raise(File_not_found e_filename)
	   | Some dirname ->
	       Godi_file.concat (getcwd()) dirname in
       let parsedir_s = Godi_file.slashname parsedir in
       let parsefile = Godi_file.addname parsedir basename in
       let parsefile_s = Godi_file.slashname parsefile in
       try
	 let r = 
	   ( try
	       parse_fd e_filename fd 
	     with
	       | e ->
		   prerr_endline ("Exception " ^ Printexc.to_string e);
		   raise e
	   )
	 in
	 close_in fd;
	 let unknown = { loc_filename = "<implicit>"; loc_linenumber = 0 } in
	 
	 [ unknown, M_push (S.mk ".PARSEDIR");
	   unknown, M_push (S.mk ".PARSEFILE");
	   unknown, M_push (S.mk ".SYSTEM_SCOPE");
	   unknown, M_assign_exp (LHS_val(V_lit ".PARSEDIR"),  V_lit parsedir_s);
	   unknown, M_assign_exp (LHS_val(V_lit ".PARSEFILE"), V_lit parsefile_s);
	   unknown, (if system_scope then
		       M_assign_exp (LHS_val(V_lit ".SYSTEM_SCOPE"),  
				     V_lit "1")
		     else
		       M_undef (S.mk ".SYSTEM_SCOPE"));
	 ]
	 @ r @
	 [ unknown, M_pop;
	   unknown, M_pop;
	 ]
       with
	   error ->
	     close_in fd;
	     raise error
    )
;;

(**********************************************************************)
(* The evaluator                                                      *)
(**********************************************************************)


let no_magic =
  { magic_target = None;
    magic_target_suffix = None;
    magic_sources = [];
    magic_sources_ood = [];
    magic_exec_level = None;
  }
;;


let var_value ~magic_context loc =
  function
      V_normal v -> v
    | V_env v -> v
    | V_override v -> v
    | V_undef -> V_lit ""
    | V_magic f -> f loc magic_context
;;


let remove_empty_words =
  List.filter (fun word -> word <> "")
;;


let suffix_split_re =
  (* Used for :E and :R modifiers.
   * Difference to bmake: '/' cannot be part of a suffix CHECK
   *)
  Str.regexp "^\\(.*\\)\\.\\([^./]\\)$";;

let component_split_re =
  (* Used for :H and :T modifiers *)
  Str.regexp "^\\(.*\\)/\\([^/]*\\)$";;


let sh_quote = Godi_match.quote_shell_word

let to_lower s =
  (* Unlike String.lowercase only convert 7 bit letters *)
  let s = String.copy s in
  for k = 0 to String.length s - 1 do
    match s.[k] with
	'A'..'Z' ->
	  s.[k] <- Char.chr (Char.code s.[k] + 32)
      | _ ->
	  ()
  done;
  s
;;


let to_upper s =
  (* Unlike String.uppercase only convert 7 bit letters *)
  let s = String.copy s in
  for k = 0 to String.length s - 1 do
    match s.[k] with
	'a'..'z' ->
	  s.[k] <- Char.chr (Char.code s.[k] - 32)
      | _ ->
	  ()
  done;
  s
;;


let prepare_substitution loc lhs rhs =
  (* Compute the regexp and the replacement for the "=" modifier *)
  (* Cases:
   * - neither lhs nor rhs contain '%': This is a traditional
   *   suffix-only substitution
   * - lhs does not contain '%', but rhs does: IMO, this does not
   *   make sense. The '%' in the rhs remains literally.
   *   (GNU make: same; bmake: strange effects)
   * - lhs contains '%', but rhs does not contain '%': This means
   *   that the words selected by the lhs are completely replaced by
   *   the constant rhs.
   *   (GNU make: same; bmake: strange effects)
   * - both lhs and rhs contain '%': The characters matched by '%' on the
   *   lhs are copied to the rhs as substitute for '%'.
   *)
  try
    Godi_match.parse_percent_subst_expr
      ~indicate_no_percent:true
      (lhs,rhs)
  with
      Godi_match.No_percent_symbol ->
	Godi_match.parse_suffix_subst_expr (lhs,rhs)
    | Godi_match.Bad_subst_expr ->
	raise(Eval_error(loc, "Bad substitution: " ^ lhs ^ "=" ^ rhs))
;;


let rec expand ?(pref_user_error=false) 
               ~magic_context ?(enable_env_check = true) 
               ctx loc (value:value) : string =

  let match_re value =
    try
      let re_str = Godi_match.regexp_of_glob_expr
		     ~protect_period:false
		     ~protect_slash:false
		     (Godi_match.parse_glob_expr
			(expand ~magic_context ~enable_env_check ctx loc value)) in
      Str.regexp re_str
    with
	Godi_match.Bad_glob_expr s ->
	  raise(Eval_error(loc, ("Bad glob expression: " ^ s)))
      | Godi_match.Unsupported_expr s ->
	  raise(Eval_error(loc, ("Unsupported glob expression: " ^ s)))
  in

  let apply_subst subst sm s =
    (* Apply the subsititution [subst] according to the modifiers [sm] to
     * the words of the string [s], and return the resulting words as
     * space-separated string.
     *)
    let words = split_in_words s in
    let first = ref true in
    let words' =
      List.map
	(fun word ->
	   let word' =
	     if !first || not sm.sm_first then
	       Godi_match.substitute ~globally:sm.sm_global subst word
	     else
	       word in
	   first := false;
	   word'
	)
	words in
    String.concat " " words'
  in

  let rec inner_expand loop_vars value =
    match value with
	V_list l ->
	  let l' = List.map (inner_expand loop_vars) l in
	  String.concat "" l'
      | V_lit s ->
	  s
      | V_esc c ->
	  "\\" ^ String.make 1 c
      | V_var(vname,mods) ->
	  let name = expand ~magic_context ~enable_env_check ctx loc vname in
	  let hname = S.mk name in
	  if SSet.mem hname loop_vars then
	    raise(Eval_error(loc, ("Variable references itself: " ^ name)));
	  let vval = hfind_var ~enable_env_check ctx hname in
	  let defined = vval <> V_undef in
	  let s =
	    inner_expand
	      (SSet.add hname loop_vars)
	      (var_value ~magic_context loc vval) in
	  snd
	    (List.fold_left
	       (fun (defined,s) m ->
		  (true, apply_modifier name (defined,s) m))
	       (defined,s)
	       mods
	    )
      | V_hashed_var(hname,mods) ->
	  let name = S.dest hname in
	  if SSet.mem hname loop_vars then
	    raise(Eval_error(loc, ("Variable references itself: " ^ name)));
	  let vval = hfind_var ~enable_env_check ctx hname in
	  let defined = vval <> V_undef in
	  let s =
	    inner_expand
	      (SSet.add hname loop_vars)
	      (var_value ~magic_context loc vval) in
	  snd
	    (List.fold_left
	       (fun (defined,s) m ->
		  (true, apply_modifier name (defined,s) m))
	       (defined,s)
	       mods
	    )
      | V_replaced_var(name, s, mods) ->
	  (List.fold_left
	     (fun s m ->
		apply_modifier name (true,s) m)
	     s
	     mods
	  )


  and apply_modifier var_name (defined,s) m =
    (* var_name: The name of the variable to which the modifier is applied
     * defined: whether the variable was defined (undefined implies s = "")
     *   Note: after the first call, the value is always considered as a
     *   defined value (see the List.fold_left loop above)
     * s: the string value of the variable
     * Returns: the modified string value
     *)
    match m with
	MO_E ->
	  (* Replace each word with its suffix; remove words without suffix *)
	  let words = split_in_words s in
	  let words' =
	    List.map
	      (fun word ->
		 if Str.string_match suffix_split_re word 0 then
		   Str.matched_group 2 word
		 else
		   "")
	      words in
	  String.concat " " (remove_empty_words words')
      | MO_R ->
	  (* Replace each word with everything but its suffix *)
	  let words = split_in_words s in
	  let words' =
	    List.map
	      (fun word ->
		 if Str.string_match suffix_split_re word 0 then
		   Str.matched_group 1 word
		 else
		   word)
	      words in
	  String.concat " " (remove_empty_words words')
      | MO_T ->
	  (* Replace each word with the last component.
	   * Very similar to "basename", but not identical; especially:
	   * x => x     ( no slash: return x )
	   * x/y => y   ( everything after "/" )
	   * x/  => ""  ( everything after "/" )
	   * /x  => x   ( everything after "/" )
	   *)
	  let words = split_in_words s in
	  let words' =
	    List.map
	      (fun word ->
		 if Str.string_match component_split_re word 0 then
		   Str.matched_group 2 word
		 else
		   word)
	      words in
	  String.concat " " (remove_empty_words words')
      | MO_H ->
	  (* Replace each word with everything but the last component.
	   * Very similar to "dirname", but not identical; especially:
	   * x => .     ( no slash: return "." )
	   * x/y => x   ( everything before "/" )
	   * x/  => x   ( everything before "/" )
	   * /x  => ""  ( everything before "/" )
	   *)
	  let words = split_in_words s in
	  let words' =
	    List.map
	      (fun word ->
		 if Str.string_match component_split_re word 0 then
		   Str.matched_group 1 word
		 else
		   ".")
	      words in
	  String.concat " " (remove_empty_words words')
      | MO_M value ->
	  (* Select only the words that match the glob expression [value] *)
	  let re = match_re value in
	  let words = split_in_words s in
	  let words' =
	    List.filter
	      (fun word -> Str.string_match re word 0)
	      words in
	  String.concat " " (remove_empty_words words')
      | MO_N value ->
	  (* Select only the words that do not match the glob expression
	   * [value]
	   *)
	  let re = match_re value in
	  let words = split_in_words s in
	  let words' =
	    List.filter
	      (fun word -> not (Str.string_match re word 0))
	      words in
	  String.concat " " (remove_empty_words words')
      | MO_native ->
	  (* Transform filenames to native format: *)
	  let fn = Godi_file.filename s in
	  Godi_file.nativename fn
      | MO_O ->
	  (* Order the words alphabetically. We assume the POSIX locale here *)
	  let words = split_in_words s in
	  let words' = List.sort compare words in
	  String.concat " " words'
      | MO_Q ->
	  (* Quote shell metacharacters with backslash *)
	  sh_quote s
      | MO_tl ->
	  to_lower s
      | MO_tu ->
	  to_upper s
      | MO_S (lhs, rhs, sm) ->
	  let lhs_s = expand ~magic_context ~enable_env_check ctx loc lhs in
	  let rhs_s = expand ~magic_context ~enable_env_check ctx loc rhs in
	  let subst = Godi_match.parse_substring_subst_expr (lhs_s,rhs_s) in
	  (* CHECK: exceptions *)
	  apply_subst subst sm s
      | MO_C (lhs, rhs, sm) ->
	  let lhs_s = expand ~magic_context ~enable_env_check ctx loc lhs in
	  let rhs_s = expand ~magic_context ~enable_env_check ctx loc rhs in
	  ( try
	      let lhs_re = Godi_match.extended_regexp lhs_s in
	      let subst = Godi_match.create_subst_expr lhs_re rhs_s in
	      apply_subst subst sm s
	    with
		Godi_match.Bad_regular_expr s ->
		  raise(Eval_error(loc, ("Bad regular expression: " ^ s)))
	      | Godi_match.Unsupported_expr s ->
		  raise(Eval_error(loc, ("Unsupported regular expression: " ^ s)))
	  )
      | MO_u ->
	  (* "uniq": Remove adjacent identical words *)
	  let words = split_in_words s in
	  let words' =
	    List.fold_left
	      (fun acc word ->
		 match acc with
		     last_word :: acc' ->
		       if word = last_word then acc else word :: acc
		   | [] ->
		       [ word ])
	      []
	      words in
	  String.concat " " (List.rev words')

      | MO_if (true_val, false_val) ->     (* ? str : str *)   (* TOTEST *)
	  (* The value of the variable is evaluated as or-expression, and
	   * if true, true_val is the result, otherwise false_val.
	   * Parsing errors count as "otherwise".
	   *)
	  ( try
	      let (e,k) = parse_or_expr (fun _ -> parse_error "dummy") s 0 in
	      if k < String.length s then parse_error "dummy";
	      let b = eval_expr ~magic_context ~enable_env_check ctx loc e in
	      if b then
		expand ~magic_context ~enable_env_check ctx loc true_val
	      else
		expand ~magic_context ~enable_env_check ctx loc false_val
	    with
		Parse_error(_,_) ->
		  expand ~magic_context ~enable_env_check ctx loc false_val
	  )

      | MO_subst (lhs_val, rhs_val) ->   (* str=str *)
	  let lhs = expand ~magic_context ~enable_env_check ctx loc lhs_val in
	  let rhs = expand ~magic_context ~enable_env_check ctx loc rhs_val in
	  let subst_expr = prepare_substitution loc lhs rhs in
	  let words = split_in_words s in
	  let words' =
	    List.map
	      (fun word ->
		 Godi_match.substitute subst_expr word
	      )
	      words in
	  String.concat " " (remove_empty_words words')
      | MO_U value ->
	  (* If the variable is undefined, [value] is the substitute *)
	  if not defined then
	    expand ~magic_context ~enable_env_check ctx loc value
	  else
	    s
      | MO_D value ->
	  (* If the variable is defined, [value] is the substitute *)
	  if defined then
	    expand ~magic_context ~enable_env_check ctx loc value
	  else
	    s   (* should be "" *)
      | MO_L ->
	  var_name
      | MO_P ->
	  (* This is always var_name because we do not support search paths! *)
	  var_name
      | MO_cmd value ->             (* !cmd! *)
	  let cmd_str = expand ~magic_context ~enable_env_check ctx loc value in
	  let cmd = command ~magic_context ctx cmd_str in
	  let (r,p) = ctx.capture_sh_cmd ctx cmd in
	  check_pstatus loc cmd_str p;
	  let r' = Str.replace_first sh_trailing_newline_re "" r in
	  let r'' = Str.global_replace sh_newline_re " " r' in
	  r''
      | MO_sh ->                        (* sh *)
	  let cmd = command ~magic_context ctx s in
	  let (r,p) = ctx.capture_sh_cmd ctx cmd in
	  check_pstatus loc s p;
	  let r' = Str.replace_first sh_trailing_newline_re "" r in
	  let r'' = Str.global_replace sh_newline_re " " r' in
	  r''
      | MO_map (var_val, str_val) ->   (* @name@ *)
	  let var_name = expand ~magic_context ~enable_env_check ctx loc var_val in
	  let words = split_in_words s in
	  let old_var_val = find_var ctx var_name in
	  (match old_var_val with
	       V_magic _ ->
		 raise(Eval_error(loc, "Attempt to set the magic variable " ^ var_name))
	     | _ -> ()
	  );
	  let strs =
	    List.map
	      (fun word ->
		 (* Assign var_name to word, and evaluate str_val *)
		 H.replace ctx.vars (S.mk var_name) (V_normal(V_lit word));
		 expand ~magic_context ~enable_env_check ctx loc str_val
	      )
	      words in
	  H.replace ctx.vars (S.mk var_name) old_var_val;
	  String.concat " " strs
  in

  catch_user_error
    ~pref_user_error
    (inner_expand SSet.empty) value


and eval_expr ~magic_context ?enable_env_check ctx loc (e:expr) : bool =
  let f_of_s x =
    try float_of_string x 
    with
      Failure s -> 
	  raise(Eval_error(loc, "Not a numeric literal: " ^ x))
  in

  let cmp_num op v1 v2 =
    let s1 = expand ~magic_context ?enable_env_check ctx loc v1 in
    let s2 = expand ~magic_context ?enable_env_check ctx loc v2 in
    let f1 = f_of_s s1 in
    let f2 = f_of_s s2 in
    op f1 f2
  in

  match e with
      E_or l ->
	List.exists (eval_expr ~magic_context ?enable_env_check ctx loc) l
    | E_and l ->
	List.for_all (eval_expr ~magic_context ?enable_env_check ctx loc) l
    | E_not g ->
	not(eval_expr ~magic_context ?enable_env_check ctx loc g)
    | E_defined v ->
	let name = expand ~magic_context ?enable_env_check ctx loc v in
	find_var ?enable_env_check ctx name <> V_undef
    | E_make v ->
	let name = expand ~magic_context ?enable_env_check ctx loc v in
	(List.mem name ctx.final_targets ||
	 (ctx.final_targets = [] && ctx.default_target = Some name))
    | E_exists v ->
	let name = expand ~magic_context ?enable_env_check ctx loc v in
	let fullname = Godi_file.concat ctx.objdir (Godi_file.filename name) in
	(* prerr_endline ("Checking file " ^ fullname); *)
	Godi_file.file_exists fullname
	  (* We do NOT search the file; bmake allows this *)
    | E_empty var ->
	let v = expand ~magic_context ?enable_env_check ctx loc (V_var var) in
	Str.string_match empty_re v 0
    | E_target v ->
	let name = expand ~magic_context ?enable_env_check ctx loc v in
	H.mem ctx.rules (S.mk name)
    | E_commands v ->
	let name = expand ~magic_context ?enable_env_check ctx loc v in
	( try
	    let rules = H.find ctx.rules (S.mk name) in
	    List.exists (fun r -> r.rule_commands <> []) rules
	  with
	      Not_found -> false
	)
    | E_num_eq (v1, v2) ->
	cmp_num ( = ) v1 v2
    | E_num_ne (v1, v2) ->
	cmp_num ( <> ) v1 v2
    | E_num_le (v1, v2) ->
	cmp_num ( <= ) v1 v2
    | E_num_lt (v1, v2) ->
	cmp_num ( < ) v1 v2
    | E_num_ge (v1, v2) ->
	cmp_num ( >= ) v1 v2
    | E_num_gt (v1, v2) ->
	cmp_num ( > ) v1 v2
    | E_str_eq (v1, v2) ->
	let s1 = expand ~magic_context ?enable_env_check ctx loc v1 in
	let s2 = expand ~magic_context ?enable_env_check ctx loc v2 in
	s1 = s2
    | E_str_ne (v1, v2) ->
	let s1 = expand ~magic_context ?enable_env_check ctx loc v1 in
	let s2 = expand ~magic_context ?enable_env_check ctx loc v2 in
	s1 <> s2


and expand_var ?pref_user_error ~magic_context ?enable_env_check ctx loc name =
  expand 
    ?pref_user_error ~magic_context ?enable_env_check 
    ctx loc (V_var(V_lit name, []))

and expand_hashed_var ~magic_context ?enable_env_check ctx loc hname =
  expand ~magic_context ?enable_env_check ctx loc (V_hashed_var(hname, []))

and command ~magic_context ctx cmd_str =
  { command = cmd_str;
    directory = Some ctx.objdir;
    env = ( List.flatten
	      (List.map
		 (fun varname ->
		    try
		      [ let varvalue =
			  expand_var 
			    ~magic_context
			    ~pref_user_error:true
			    ~enable_env_check:false ctx loc_unknown varname in
			(varname, varvalue)
		      ]
		    with
		      | User_error msg ->
			  printmsg ctx.console
			    (sprintf "While exporting the variable '%s': %s\n"
			       varname msg);
			  []
		 )
		 ctx.exported_vars) 
	    @ [ "PWD", (Godi_file.slashname ctx.objdir) ]
	  );
  }

and exists_var ctx name =
  try
    H.find ctx.vars (S.mk name) <> V_undef
  with
      Not_found -> false

and find_var ?(enable_env_check = true) ctx name =
  try
    let r = H.find ctx.vars (S.mk name) in
    if enable_env_check && ctx.warn_on_env_access then
      (match r with
	   V_env _ ->
	     maybe_env_warning ctx name
	 | _ -> ()
      );
    r
  with
      Not_found -> 
	if enable_env_check && ctx.warn_on_env_access then
	  maybe_env_warning ctx name;
	V_undef

and hfind_var ?(enable_env_check = true) ctx hname =
  try
    let r = H.find ctx.vars hname in
    if enable_env_check && ctx.warn_on_env_access then
      (match r with
	   V_env _ ->
	     maybe_env_warning ctx (S.dest hname)
	 | _ -> ()
      );
    r
  with
      Not_found -> 
	if enable_env_check && ctx.warn_on_env_access then
	  maybe_env_warning ctx (S.dest hname);
	V_undef

and maybe_env_warning ctx name =
  (* If name is listed in GODI_MAKE_ENV, the access is registered, and
   * no warning is printed. Otherwise, output to console.
   *)
  if name <> "GODI_MAKE_ENV" then (
    let godi_make_env = 
      expand_var
	~enable_env_check:false ~magic_context:no_magic 
	ctx loc_unknown "GODI_MAKE_ENV" in
    (* Note: No env check during the expansion of GODI_MAKE_ENV; this could
     * lead to an endless loop.
     *)
    let words = split_in_words godi_make_env in
    if not (List.mem name words) then
      printmsg ctx.console
	(sprintf 
	   "WARNING: Variable `%s' is uninitialized; this might access the environment unintentionally\n"
	   name)
  )
;;


let rec expand_only ~magic_context ctx vars loc v =
  (* Expands only the variables listed in [vars], the other variables remain
   * unexpanded. Variables with computed names remain unexpanded, too.
   *)
 ( match v with
       V_lit _  -> v
     | V_esc _  -> v
     | V_list l -> V_list(List.map(expand_only ~magic_context ctx vars loc) l)
     | V_var(v',mods) ->
	 let new_mods = 
	   List.map (expand_only_mods ~magic_context ctx vars loc) mods in
	 ( try
	     let name = const_value v' in  (* or Not_found *)
	     if List.mem name vars then
	       let s = expand ~magic_context ctx loc v in
	       V_replaced_var (name, s, new_mods)
	     else
	       (V_var(v', new_mods))
	   with
	       Not_found ->
		 let new_v' = expand_only ~magic_context ctx vars loc v' in
		 V_var(new_v', new_mods)
	 )
	   
     | V_hashed_var(v',mods) ->
	 let new_mods = 
	   List.map (expand_only_mods ~magic_context ctx vars loc) mods in
	 let name = S.dest v' in
	 if List.mem name vars then
	   let s = expand_hashed_var ~magic_context ctx loc v' in
	   V_replaced_var (name, s, new_mods)
	 else
	   V_hashed_var(v', new_mods)
	     
     | V_replaced_var (name, s, mods) ->
	 (* The variable is not replaced again, but the modifiers may 
          * be updated
          *)
	 let new_mods = 
	   List.map (expand_only_mods ~magic_context ctx vars loc) mods in
	 V_replaced_var (name, s, new_mods)
 )

and expand_only_mods ~magic_context ctx vars loc m =
  match m with
    | MO_E -> MO_E
    | MO_H -> MO_H
    | MO_M v -> 
	MO_M (expand_only ~magic_context ctx vars loc v)
    | MO_N v -> 
	MO_N (expand_only ~magic_context ctx vars loc v)
    | MO_O -> MO_O
    | MO_Q -> MO_Q
    | MO_R -> MO_R
    | MO_tl -> MO_tl
    | MO_tu -> MO_tu
    | MO_S (v1,v2,sm) -> 
	MO_S (expand_only ~magic_context ctx vars loc v1,
	      expand_only ~magic_context ctx vars loc v2,
	      sm)
    | MO_C (v1,v2,sm) -> 
	MO_C (expand_only ~magic_context ctx vars loc v1,
	      expand_only ~magic_context ctx vars loc v2,
	      sm)
    | MO_T -> MO_T
    | MO_u -> MO_u
    | MO_if (v1,v2) -> 
	MO_if (expand_only ~magic_context ctx vars loc v1,
	       expand_only ~magic_context ctx vars loc v2)
    | MO_subst (v1,v2) -> 
	MO_subst (expand_only ~magic_context ctx vars loc v1,
		  expand_only ~magic_context ctx vars loc v2)
    | MO_U v -> 
	MO_U (expand_only ~magic_context ctx vars loc v)
    | MO_D v -> 
	MO_D (expand_only ~magic_context ctx vars loc v)
    | MO_L -> MO_L
    | MO_P -> MO_P
    | MO_cmd v -> 
	MO_cmd (expand_only ~magic_context ctx vars loc v)
    | MO_sh -> MO_sh
    | MO_map (v1,v2) -> 
	MO_map (expand_only ~magic_context ctx vars loc v1,
		expand_only ~magic_context ctx vars loc v2)
;;


let search_user_file ctx file =
  let found_file =
    search_file ctx.objdir (ctx.user_mk_path @ ctx.sys_mk_path) file in
  parse found_file
;;


let search_sys_file ctx file =
  let found_file =
    search_file ctx.objdir ctx.sys_mk_path file in
  parse ~system_scope:true found_file
;;


let null_rule = 
  { rule_target = "";
    rule_target_suffix = "";
    rule_sources = [];
    rule_atts = [];
    rule_commands = [];
    rule_type = `Dep;
    rule_file_type = `File;
    rule_ignore_flag = false;
    rule_made_flag = false;
    rule_force_exec_flag = false;
    rule_optional_flag = false;
    rule_silent_flag = false;
    rule_system_flag = false;
  }
;;


let eval ?(pref_user_error = false)
         ?(include_user_file = search_user_file)
         ?(include_sys_file = search_sys_file)
	 ctx cmds =

  let var_stack = Stack.create() in
    (* for push/pop *)

  let lhs_hname loc lhs =
    match lhs with
	LHS_val v ->
	  S.mk(expand ~magic_context:no_magic ctx loc v)
      | LHS_hashed_var hname ->
	  hname
  in

  let replace ?(force=false) loc hname entry =
    (* Replace the entry for the variable [hname] by [entry]. If [force=false],
     * V_override entries are not replaced.
     *)
    let old_value = hfind_var ~enable_env_check:false ctx hname in
    let do_replace =
      match old_value with
	  V_override _ ->
	    force
	| V_magic _ ->
	    raise(Eval_error(loc, "Attempt to set the magic variable " ^ 
			       S.dest hname))
	| _ ->
	    true
    in
    if do_replace then
      H.replace ctx.vars hname entry
  in
  let replace_value ?force loc lhs value =
    (* Replace the entry for the variable [lhs] by [value]. *)
    let hname = lhs_hname loc lhs in
    replace ?force loc hname (V_normal value)
  in   

  let process_special_sources loc rule =
    (* Checks for special sources in rule.rule_sources, and sets
     * rule_sources, rule_atts, and the flags accordingly.
     *)
    (* First split into regular and special sources: *)
    let is_special src =
      (* CHECK: This test is questionable. *)
      Str.string_match special_source_re src 0 in
    let special_srcs, regular_srcs =
      List.partition is_special rule.rule_sources in
    (* Set the rule components: *)
    rule.rule_sources <- regular_srcs;
    rule.rule_atts <- special_srcs;
    let have_notmain = ref false in
    let have_use = ref false in  (* .USE or .USEBEFORE *)
    List.iter
      (function
	   ".IGNORE"   -> rule.rule_ignore_flag <- true
	 | ".MADE"     -> rule.rule_made_flag <- true
	 | ".MAKE"     -> rule.rule_force_exec_flag <- true
	 | ".OPTIONAL" -> rule.rule_optional_flag <- true
	 | ".PRECIOUS" -> ctx.precious_targets <- 
	                      rule.rule_target :: ctx.precious_targets
	 | ".SILENT"   -> rule.rule_silent_flag <- true
	 | ".JOIN"     -> rule.rule_file_type <- `Join
                          (* may be overridden by a .PHONY decl *)
	 | ".NOTMAIN"  -> have_notmain := true;
	 | ".USE" 
	 | ".USEBEFORE" -> have_use := true
	 | _ -> ()
      )
      special_srcs;
    (* Is this the default rule? *)
    let target_is_special =
      rule.rule_target <> "" && rule.rule_target.[0] = '.' in
    if ctx.default_target = None && not !have_notmain &&
       not target_is_special
    then
      ctx.default_target <- Some rule.rule_target;
    (* .USE, .USEBEFORE trigger errors *)
    if !have_use then
      raise(Eval_error(loc, ".USE macros are not implemented"));
    (* Note: .WAIT is silently ignore *)
  in

  let process_regular_dep typ loc target sources shell_cmds =
    (* Enters a regular dependency into ctx.rules. [typ] is the type of
     * the dependency (which operator is used: ":" or "::" or "!").
     * [loc] is the location of the dependency line. [target] is a
     * single, already expanded target (left side). [sources] is the
     * list of already expanded sources (right side). [shell_cmds]
     * are the shell commands following the dependency line, with the
     * same type as [rule_commands].
     *)
    let target_suffix = default_target_suffix target in
    ( try
	let rules = H.find ctx.rules (S.mk target) in  (* or Not_found *)
	let rule = List.hd rules in
	if rule.rule_type <> typ then
	  raise(Eval_error(loc, ("Inconsistent dependency operator for the target `" ^ target ^ "'")));
	( match typ with
	      (`Dep|`Forced_dep) ->
		(* Accumulate sources. Only the first script counts *)
		rule.rule_sources <- rule.rule_sources @ sources;
		if rule.rule_commands <> [] && shell_cmds <> [] &&
		  not rule.rule_system_flag
		then
		  raise(Eval_error(loc, ("Duplicate script for target `" ^ target ^ "'")));
		process_special_sources loc rule;
		rule.rule_commands <- 
		  if shell_cmds <> [] then shell_cmds else rule.rule_commands
	    | `Concat_dep ->
		(* Accumulate independent rules *)
		let rule' =
		  { null_rule with
		      rule_target = target;
		      rule_target_suffix = target_suffix;
		      rule_sources = sources;
		      rule_commands = shell_cmds;
		      rule_type = typ;
		  } in
		process_special_sources loc rule';
		H.replace ctx.rules (S.mk target) (rules @ [rule']);
	);
      with
	  Not_found ->
	    let sysflag = exists_var ctx ".SYSTEM_SCOPE" in
	    let rule =
	      { null_rule with
		  rule_target = target;
		  rule_target_suffix = target_suffix;
		  rule_sources = sources;
		  rule_commands = shell_cmds;
		  rule_type = typ;
		  rule_system_flag = sysflag;
	      } in
	    process_special_sources loc rule;
	    H.replace ctx.rules (S.mk target) [rule];
    )
  in

  let process_single_dep typ loc target sources shell_cmds =
    (* First check whether we have a special target. If not, go on with
     * [process_regular_dep]. Else call the special processor for the
     * target.
     *)
    if String.length target > 0 && target.[0] = '.' then (
      match target with
	  (* --- Implemented targets: --- *)
	  ".BEGIN"
	| ".END"
	| ".INTERRUPT" ->
	    (* These three targets are propagated to "::" style: *)
	    process_regular_dep `Concat_dep loc target sources shell_cmds

	| ".IGNORE"
	| ".PRECIOUS"
	| ".SILENT"
	| ".PHONY" ->
	    (* These targets must not have shell commands. They are propagated
	     * to ":" style if sources <> []. For empty sources, the global
	     * flag is set, if any.
	     *)
	    if shell_cmds <> [] then
	      raise(Eval_error(loc, "This target must not have shell commands"));
	    if sources = [] then
	      ( match target with
		    ".IGNORE"   -> ctx.global_ignore_flag <- true
		  | ".PRECIOUS" -> ctx.global_precious_flag <- true
		  | ".SILENT"   -> ctx.global_silent_flag <- true
		  | _           -> ()
	      )
	    else
	      process_regular_dep `Dep loc target sources []
	    
	| ".MAIN" ->
	    if List.length sources <> 1 then
	      raise(Eval_error(loc, ".MAIN must have exactly one argument"));
	    ctx.default_target <- Some(List.hd sources);

	| ".SUFFIXES" ->
	    (* Empty list of sources deletes the list of suffixes! *)
	    if shell_cmds <> [] then
	      raise(Eval_error(loc, "This target must not have shell commands"));
	    if sources = [] then
	      ctx.suffixes <- []
	    else
	      ctx.suffixes <- ctx.suffixes @ sources
	    
	  (* --- Targets going to be implemented: --- *)
	| ".ORDER" ->
	    ()   (* Defines additional dependencies. We need the graph first *)
	| ".MAKEFLAGS" | ".MFLAGS" ->
	    ()   (* We do not have the parser for arguments yet *)

          (* --- Ignored unimplemented targets: --- *)
	| ".INCLUDES"
	| ".LIBS"
	| ".NOPATH"
	| ".NOTPARALLEL"
	| ".NO_PARALLEL"
	| ".NULL"
	| ".PATH"
	| ".POSIX"
	| ".SHELL"
	| ".SINGLESHELL"
	| ".WAIT" -> ()

	  (* --- Unimplemented targets raising errors: --- *)
	| ".DEFAULT" ->
	    raise(Eval_error(loc, ".DEFAULT is not implemented"))

	| _ ->
	    (* This may also be an implicit rule, e.g. ".c.o", so we handle
             * this case as regular case 
             *)
	    process_regular_dep typ loc target sources shell_cmds
    )
    else
      process_regular_dep typ loc target sources shell_cmds
  in

  let process_dep typ loc target_val source_val =
    (* Enters a dependency line into ctx.rules. [typ] and [loc] see
     * [process_single_dep] above. [target] is the unexpanded target
     * as string. [source] is the unexpanded source as string.
     * The function returns a list of functions which must still be
     * applied to [shell_cmds] in order to get executed. Only when
     * these functions are executed, the dependency are really entered.
     * [shell_cmds] are the script commands following the dependency line.
     * The function has the side effect of changing ctx.rules.
     *)
    (*
     * - [target] may contain wildcards in the last component of the
     *   file path. Resolve these, and split into space-separated words.
     *   Simply iterate over the resulting targets, and enter them
     *   individually into the table.
     * - [source] same
     *)
    let base_dir = ctx.objdir in
    let targets = generate_files
		    ~base_dir loc (expand 
				     ~magic_context:no_magic ctx loc target_val)
    in
    let processor = ref [] in
    List.iter
      (fun t ->
	 let t_suffix = default_target_suffix t in
	 let magic_context =
	   { no_magic with 
	       magic_target = Some t;
	       magic_target_suffix = Some t_suffix;
	   } in
	 let source_str = expand ~magic_context ctx loc source_val in
	 let sources = generate_files ~base_dir loc source_str in
	 processor := (process_single_dep typ loc t sources) :: !processor;
      )
      targets;
    !processor
  in

  let postprocess_flags() =
    (* Check .PHONY, .SILENT, .IGNORE, and .PRECIOUS targets, and set the
     * flags of the mentioned rules
     *)
    let get_sources target =
      try
	let rules = H.find ctx.rules (S.mk target) in
	H.remove ctx.rules (S.mk target);
	List.flatten (List.map (fun rule -> rule.rule_sources) rules)
      with
	  Not_found -> [] 
    in

    let apply_to_rule f targets =
      List.iter
	(fun target ->
	   let rules = try H.find ctx.rules (S.mk target) with Not_found -> []
	   in
	   List.iter f rules
	)
	targets
    in

    let phony_targets = get_sources ".PHONY" in
    let silent_targets = get_sources ".SILENT" in
    let ignored_targets = get_sources ".IGNORE" in
    let precious_targets = get_sources ".PRECIOUS" in

    apply_to_rule (fun r -> r.rule_file_type <- `Phony) phony_targets;
    apply_to_rule (fun r -> r.rule_silent_flag <- true) silent_targets;
    apply_to_rule (fun r -> r.rule_ignore_flag <- true) ignored_targets;
    ctx.precious_targets <- ctx.precious_targets @ precious_targets;
  in

  let current_dep = ref None in
  let current_shell_cmds_rev = ref [] in

  let handle_dep() =
    match !current_dep with
	Some processor ->
	  let shell_cmds = List.rev !current_shell_cmds_rev in
	  List.iter (fun proc_item -> proc_item shell_cmds) processor;
	  current_dep := None;
	  current_shell_cmds_rev := [];
      | None ->
	  ()
  in

  let rec eval_loop for_vars cmds =
    (* for_vars: The list of variables with ".for" semantics, i.e. these
     * are immediately expanded in this loop
     *)
    let expand_for_vars loc v =
      if for_vars = [] then 
	v 
      else 
	expand_only ~magic_context:no_magic ctx for_vars loc v in

(*
    if cmds <> [] then (
      let _loc = fst (List.hd cmds) in
      prerr_endline ("EVAL file=" ^ _loc.loc_filename ^ " line=" ^
		     string_of_int _loc.loc_linenumber);
    );
*)
    match cmds with
	(loc, M_assign(lhs,value)) :: cmds' ->
	  handle_dep();
	  let value = expand_for_vars loc value in
	  replace_value loc lhs value;
	  eval_loop for_vars cmds'
      | (loc, M_assign_app(lhs,value)) :: cmds'  ->
	  handle_dep();
	  let value = expand_for_vars loc value in
	  let hname = lhs_hname loc lhs in
	  let old_value = hfind_var ctx hname in
	  let new_value =
	    if old_value <> V_undef then
	      V_list [var_value ~magic_context:no_magic loc old_value; V_lit " "; value]
		(* CHECK: no magic_context *)
	    else
	      value in
	  replace_value loc lhs new_value;
	  eval_loop for_vars cmds'
      | (loc, M_assign_opt(lhs,value)) :: cmds'  ->
	  handle_dep();
	  let value = expand_for_vars loc value in
	  let hname = lhs_hname loc lhs in
	  if hfind_var ctx hname = V_undef then
	    replace_value loc lhs value;
	  eval_loop for_vars cmds'
      | (loc, M_assign_exp(lhs,value)) :: cmds'  ->
	  handle_dep();
	  let s = expand ~magic_context:no_magic ctx loc value in
	  replace_value loc lhs (V_lit s);
	  eval_loop for_vars cmds'
      | (loc, M_assign_sh(lhs,value)) :: cmds'  ->
	  handle_dep();
	  let s = expand ~magic_context:no_magic ctx loc value in
	  let cmd = command ~magic_context:no_magic ctx s in
	  let (s', pstatus) = ctx.capture_sh_cmd ctx cmd in
	  check_pstatus loc s pstatus;
	  let s'' = Str.replace_first sh_trailing_newline_re "" s' in
	  let s''' = Str.global_replace sh_newline_re " " s'' in
	  replace_value loc lhs (V_lit s''');
	  eval_loop for_vars cmds'
      | (loc, M_undef hname) :: cmds' ->
	  handle_dep();
	  replace loc hname V_undef;
	  eval_loop for_vars cmds'
      | (loc, M_push hname) :: cmds' ->
	  handle_dep();
	  let var = hfind_var ~enable_env_check:false ctx hname in
	  Stack.push (hname,var) var_stack;
	  replace ~force:true loc hname V_undef;
	  (* Note: Even V_override variables are undef'ed *)
	  eval_loop for_vars cmds'
      | (loc, M_pop) :: cmds' ->
	  handle_dep();
	  let (hname,var) = Stack.pop var_stack in
	  replace ~force:true loc hname var;
	  eval_loop for_vars cmds'
      | (loc, M_if(e,then_cmds,else_cmds)) :: cmds' ->
	  let branch_cmds =
	    if eval_expr ~magic_context:no_magic ctx loc e then
	      then_cmds
	    else
	      else_cmds in
	  eval_loop for_vars branch_cmds;
	  eval_loop for_vars cmds'
      | (loc, M_for(vars, v, body_cmds)) :: cmds' ->
	  let s = expand ~magic_context:no_magic ctx loc v in
	  let words = split_in_words s in
	  let n_vars = List.length vars in
	  let n_words = List.length words in
	  assert(n_vars > 0);
	  if (n_words mod n_vars) <> 0 then
	    raise(Eval_error(loc, ".for: The number of words after 'in' is not divisible by the number of variables before 'in'"));
	  let rows = split_regularly n_vars words in
	  let old_var_vals = List.map (hfind_var ~enable_env_check:false ctx) vars in
	  (* We remember the old values of the variables in [old_var_vals],
	   * and restore them after the .for loop is done
	   *)
	  List.iter
	    (fun row ->
	       List.iter2
	         (fun hname value ->
		    replace ~force:true loc hname (V_normal (V_lit value))
		 )
	         vars
	         row;
	       eval_loop (List.map S.dest vars @ for_vars) body_cmds
	    )
	    rows;
	  (* Restore variables: *)
	  List.iter2
	    (fun hname value ->
	       replace ~force:true loc hname value)
	    vars
	    old_var_vals;
	  eval_loop for_vars cmds'

      | (loc, M_dep(target,source)) :: cmds' ->
	  handle_dep();
	  (* Important note: By calling process_dep here, it is ensured that
	   * the variables in the dep lines are expanded now, and not later
	   * when the shell commands have all been collected.
	   *)
	  let target = expand_for_vars loc target in
	  let source = expand_for_vars loc source in
	  current_dep := Some(process_dep `Dep loc target source);
	  eval_loop for_vars cmds'
      | (loc, M_forced_dep(target,source)) :: cmds' ->
	  handle_dep();
	  let target = expand_for_vars loc target in
	  let source = expand_for_vars loc source in
	  current_dep := Some(process_dep `Forced_dep loc target source);
	  eval_loop for_vars cmds'
      | (loc, M_concat_dep(target,source)) :: cmds' ->
	  handle_dep();
	  let target = expand_for_vars loc target in
	  let source = expand_for_vars loc source in
	  current_dep := Some(process_dep `Concat_dep loc target source);
	  eval_loop for_vars cmds'

      | (loc, M_shell v) :: cmds' ->
	  if !current_dep = None then
	    raise(Eval_error(loc, "This script is outside of any rule"));
	  (* Expand the ".for" variables now, the other variables remain
	   * unexpanded:
	   *)
	  let v' = expand_for_vars loc v in
	  current_shell_cmds_rev := (loc,v') :: !current_shell_cmds_rev;
	  eval_loop for_vars cmds'

      | (loc, M_include(ignore_flag, file_value)) :: cmds' ->
	  handle_dep();
	  let file = expand ~magic_context:no_magic ctx loc file_value in
	  let included_cmds =
	    try
	      include_user_file ctx file
	    with
		(File_not_found _ as error) ->
		  if ignore_flag then [] else raise error in
	  eval_loop [] included_cmds;
	  eval_loop for_vars cmds'

      | (loc, M_sys_include(ignore_flag, file_value)) :: cmds' ->
	  handle_dep();
	  let file = expand ~magic_context:no_magic ctx loc file_value in
	  let included_cmds =
	    try
	      include_sys_file ctx file
	    with
		(File_not_found _ as error) ->
		  if ignore_flag then [] else raise error in
	  eval_loop [] included_cmds;
	  eval_loop for_vars cmds'

      | [] ->
	  ()

  in

  catch_user_error
    ~pref_user_error
    (fun () ->
       eval_loop [] cmds;
       postprocess_flags()
    )
    ()
;;


let rec value_string =
  function
      V_lit s -> s
    | V_esc e -> "<ESC " ^ string_of_int(Char.code e) ^ ">"
    | V_var(vname,mods) ->
	"${" ^ value_string vname ^
	(if mods <> [] then ":<mods>" else "") ^ "}"
    | V_hashed_var(hname,mods) ->
	"${" ^ S.dest hname ^ 
	(if mods <> [] then ":<mods>" else "") ^ "}"
    | V_list l ->
	String.concat "" (List.map value_string l)
    | V_replaced_var(vname,s,[]) ->
	s
    | V_replaced_var(vname,s,mods) ->
	"<REPLACED " ^ s ^ ":<mods>>"
;;



let dump_context ?(env=false) ctx =
  prerr_endline "Variables:";
  H.iter
    (fun namep var ->
       let name = S.dest namep in
       match var with
	   V_normal v ->
	     let s = value_string v in
	     prerr_string ("  " ^ name ^ "=");
	     prerr_endline ("<normal> \"" ^ String.escaped s ^ "\"")
	 | V_override v ->
	     let s = value_string v in
	     prerr_string ("  " ^ name ^ "=");
	     prerr_endline ("<override> \"" ^ String.escaped s ^ "\"")
	 | V_env v ->
	     let s = value_string v in
	     if env then (
	       prerr_string ("  " ^ name ^ "=");
	       prerr_endline ("<envl> \"" ^ String.escaped s ^ "\"")
	     )
	 | V_undef ->
	     prerr_string ("  " ^ name ^ "=");
	     prerr_endline "<undef>";
	 | V_magic f ->
	     prerr_string ("  " ^ name ^ "=");
	     prerr_endline "<magic>";
    )
    ctx.vars;

  prerr_endline "Targets:";
  H.iter
    (fun target rules ->
       List.iter
         (fun rule ->
	    let op =
	      match rule.rule_type with
		  `Dep -> ":"
		| `Forced_dep -> "!"
		| `Concat_dep -> "::"
	    in
	    prerr_endline("  " ^ S.dest target ^ op ^
			  String.concat " " rule.rule_sources);
	    if rule.rule_commands <> [] then
	      prerr_endline "    <nonempty>";
	 )
         rules;
    )
    ctx.rules
;;


let var_definition ctx varname =
  let def = hfind_var ~enable_env_check:false ctx (S.mk varname) in
  let value = var_value ~magic_context:no_magic loc_unknown def in
    (* CHECK: no magic_context *)
  value_string value   (* TODO: Better reconstruction! *)
;;

(**********************************************************************)
(* Initialize Contexts                                                *)
(**********************************************************************)

let return_target loc mctx =
  (* Implementation of ${.TARGET} *)
  match mctx.magic_target with
      None ->
	raise(Eval_error(loc, "There is no .TARGET here"))
    | Some t ->
	V_lit t
;;


let return_prefix loc mctx =
  (* Implementation of ${.PREFIX} *)
  match mctx.magic_target, mctx.magic_target_suffix with
    | (Some t, Some suff) ->
	let chopped =
	  if Filename.check_suffix t suff then
	    Filename.chop_suffix t suff 
	  else
	    t
	in
	V_lit chopped
    | _ ->
	raise(Eval_error(loc, "There is no .PREFIX here"))
;;


let return_all_targets ctx loc mctx =
  (* Implementation of ${.TARGETS} *)
  let l = (match ctx.default_target with
	       None -> []
	     | Some t -> [t]) @ ctx.final_targets in
  V_lit(String.concat " " l)
;;


let return_sources ctx loc mctx =
  (* Implementation of ${.ALLSRC} *)
  V_lit(String.concat " " mctx.magic_sources)
;;


let return_first_source ctx loc mctx =
  (* Implementation of ${.IMPSRC} *)
  match mctx.magic_sources with
    | first :: _ ->
	V_lit first
    | [] ->
	raise(Eval_error(loc, "There is no .IMPSRC here"))
;;


let return_ood_sources ctx loc mctx =
  (* Implementation of ${.OODATE} *)
  V_lit (String.concat " " mctx.magic_sources_ood)
;;


let return_makelevel ctx loc mctx =
  (* Implementation of ${.MAKELEVEL} *)
  let s =
    match mctx.magic_exec_level with
      | None -> V_lit (try Sys.getenv "MAKELEVEL" with Not_found -> "")
      | Some level -> V_lit (string_of_int level)
  in
  s;
;;


let alias ctx realname loc mctx =
  (* The value of the variable is the value of [realname] *)
  var_value ~magic_context:mctx loc (find_var ctx realname)
;;


let compute_makeflags ctx loc mctx =
  (* MAKEFLAGS is the concatenation of $(.MAKEOPTIONS) and the assignments
   * that would set the variables mentioned in $(.MAKEOVERRIDES).
   *)
  let options = 
    expand_var
      ~magic_context:mctx ~pref_user_error:false ~enable_env_check:false 
      ctx loc ".MAKEOPTIONS" in
  let overrides =
    expand_var 
      ~magic_context:mctx ~pref_user_error:false ~enable_env_check:false
      ctx loc ".MAKEOVERRIDES" in
  let varlist = split_in_words overrides in
  let assignments =
    List.map
      (fun varname ->
	 let varvalue =
	   expand_var ~magic_context:mctx ~pref_user_error:false 
	     ~enable_env_check:false ctx loc varname in
	 sh_quote varname ^ "=" ^ sh_quote varvalue
      )
      varlist
  in
  V_lit(options ^ " " ^ String.concat " " assignments)
;;


let determine_curdir env : string =
  let cwd = getcwd() in
  let pwd = ref (Godi_file.slashname cwd) in
  Array.iter
    (fun s ->
       if Str.string_match env_re s 0 then (
	 let n = Str.matched_group 1 s in
	 let v = Str.matched_group 2 s in
	 let v_fn = Godi_file.filename v in
	 if n = "PWD" && Godi_file.same_file v_fn cwd  then
	   (* CHECK: This use of same_file is critical *)
	   pwd := v
       )
    )
    env;
  !pwd
;;


let default_sys_mk_path() =
  match Godi_boot.stage() with
    | `Stage0
    | `Stage1 ->
	[ try
	    Sys.getenv "GODI_BOOT_SYSMK"
	  with
	    | Not_found ->
		failwith "GODI bootstrap console: Neet env var GODI_BOOT_SYSMK"
	]
    | _ ->
	Str.split (Str.regexp ":") Godi_conf.sys_mk_path ;;

let create_context
       ?(env = Godi_exec.environment())
       ?(override = [])
       ?(options = "")
       ?(user_mk_path = [ Godi_file.filename "." ])
       ?(sys_mk_path = List.map Godi_file.filename (default_sys_mk_path()))
       ?(curdir = determine_curdir env)
       ?(objdir = getcwd())
       ?(make_cmd = Sys.argv.(0))
       ?(capture_sh_cmd = capture_sh_cmd)
       ?(call_sh_cmd = call_sh_cmd)
       ?(console = stderr_console())
       ?(warn_on_env_access = false)
       ?(trace = false)
       ?(trace_script = false)
       targets =
  let ctx =
    { vars = H.create 2000;
      exported_vars = [ "GODI_MAKEFLAGS"; "MAKELEVEL" ];
      rules = H.create 200;
      final_targets = targets;
      default_target = None;
      sys_mk_path = sys_mk_path;
      user_mk_path = user_mk_path;
      curdir = curdir;
      objdir = objdir;
      capture_sh_cmd = capture_sh_cmd;
      call_sh_cmd = call_sh_cmd;
      global_ignore_flag = false;
      global_silent_flag = false;
      global_precious_flag = false;
      suffixes = [];
      precious_targets = [];
      console = console;
      trace = trace;
      trace_script = trace_script;
      warn_on_env_access = warn_on_env_access;
    } in

  (match Godi_file.nameclass objdir with
     | `Relative | `Error ->
	 failwith "create_context: objdir must be absolute"
     | _ ->
	 ()
  );

  (* Put environment variables into vars: *)
  Array.iter
    (fun s ->
       if Str.string_match env_re s 0 then (
	 let n = Str.matched_group 1 s in
	 let v = Str.matched_group 2 s in
	 H.replace ctx.vars (S.mk n) (V_env(V_lit v));
       ) (* else: ignore s *)
    )
    env;
  (* Put overriden variables into vars (ok to overwrite environment): *)
  List.iter
    (fun (n,v) ->
       H.replace ctx.vars (S.mk n) (V_override(V_lit v));
    )
    override;
  (* Initialize special variables: *)
  let setvar n v =
    H.replace ctx.vars (S.mk n) v in
  setvar ".CURDIR" (V_magic (fun _ _ -> V_lit ctx.curdir));
  setvar ".OBJDIR" (V_magic (fun _ _ -> V_lit (Godi_file.slashname ctx.objdir)));
  setvar ".ARCHIVE" (V_magic (fun loc _ -> raise(Eval_error(loc, ".ARCHIVE is not supported"))));
  setvar ".MEMBER" (V_magic (fun loc _ -> raise(Eval_error(loc, ".MEMBER is not supported"))));
  setvar ".TARGET" (V_magic return_target);
  setvar ".PREFIX" (V_magic return_prefix);
  setvar ".TARGETS" (V_magic (return_all_targets ctx));
  setvar ".ALLSRC" (V_magic (return_sources ctx));
  setvar ".IMPSRC" (V_magic (return_first_source ctx));
  setvar ".OODATE" (V_magic (return_ood_sources ctx));
  setvar ".newline" (V_magic (fun _ _ -> V_lit "\n"));
  setvar ">" (V_magic (alias ctx ".ALLSRC"));
  setvar "<" (V_magic (alias ctx ".IMPSRC"));
  setvar "?" (V_magic (alias ctx ".OODATE"));
  setvar "*" (V_magic (alias ctx ".PREFIX"));
  setvar "@" (V_magic (alias ctx ".TARGET"));
  setvar ".MAKE" (V_magic (fun _ _ -> V_lit make_cmd));
  setvar "MAKE" (V_normal (V_var (V_lit ".MAKE", [])));
  (* setvar "MACHINE_ARCH" (V_normal (V_var (V_lit Godi_conf.machine_arch, []))); *)
  setvar ".MAKEFLAGS" (V_magic (compute_makeflags ctx));
  setvar "GODI_MAKEFLAGS" (V_normal (V_var (V_lit ".MAKEFLAGS", [])));
  setvar "MAKEFLAGS" (V_normal (V_lit ""));
  (* Note:
   * - Both MAKEFLAGS and GODI_MAKEFLAGS are accepted from the environment
   * - Only GODI_MAKEFLAGS is automatically exported
   * Advantage:
   * - Compatible with bsd.pkg.mk
   * - GNU make also interprets MAKEFLAGS, but in a different way.
   *)
  setvar ".MAKEOPTIONS" (V_magic (fun _ _ -> V_lit options));
  setvar ".MAKEOVERRIDES" 
    (V_normal (V_lit (String.concat " " (List.map fst override))));
  setvar ".MAKELEVEL" (V_magic (return_makelevel ctx));
  setvar "MAKELEVEL" (V_normal (V_var (V_lit ".MAKELEVEL", [])));
  setvar ".GODI_MAKE" (V_magic (fun _ _ -> V_lit "1"));
  setvar "GODI_MAKE_ENV" (V_normal (V_lit "OS unix GCC CC CFLAGS FFLAGS LDFLAGS CPPFLAGS AR ARFLAGS RANLIB AS AFLAGS COMPILE.s LINK.s COMPILE.S LINK.S DBG COMPILE.c LINK.c CXX CXXFLAGS COMPILE.cc LINK.cc  OBJC OBJCFLAGS COMPILE.m LINK.m CPP NOLINT  FC RFLAGS COMPILE.f LINK.f COMPILE.F LINK.F COMPILE.r LINK.r INSTALL LEX LFLAGS LEX.l LD LINT LINTFLAGS LORDER MAKE NM PC PFLAGS COMPILE.p LINK.p SHELL SIZE YACC YFLAGS YACC.y TSORT"));
  ctx
;;

(**********************************************************************)
(* Parser cache                                                       *)
(**********************************************************************)

type file_id = Godi_file.file_id

type file_info = Godi_file.file_fingerprint

type parser_cache = (file_id, file_info * loc_make_cmd list) Hashtbl.t
  (* Maps from file_ids to pairs of (file_info, cmdlist). The filenames
   * must be absolute.
   *)

let create_parser_cache() = Hashtbl.create 50;;

(* Minor bug: When the same file is accessed twice, but with different 
 * system_scope, the effective scope is taken from the first access.
 *)

let fast_parse_eval 
      ?(pref_user_error=false) ?(sys_mk=true) ?make_user_file
      pcache main_file ctx =

  let get_file system_scope path file =
    let abs_file = 
      catch_user_error 
	~pref_user_error 
	(fun file -> 
	   try
	     search_file ctx.objdir path file
	   with
	       File_not_found _ 
		 when not system_scope && make_user_file <> None ->
		   (match make_user_file with
		      | None -> assert false
		      | Some f -> f file ctx
		   );
		   (* Hopefully the file exists now: *)
		   search_file ctx.objdir [ Godi_file.filename "."] file
	) 
	file in
    let sid = Godi_file.file_id abs_file in
    let sinfo = Godi_file.file_fingerprint abs_file in
    try
      let (finfo,cmds) = Hashtbl.find pcache sid in
      if sinfo <> finfo then
	raise Not_found  (* force parsing *)
      else
	cmds
    with
	Not_found ->
	  (* Parse (again) *)
	  let cmds = parse ~pref_user_error ~system_scope abs_file in
	  Hashtbl.replace pcache sid (sinfo, cmds);
	  cmds
  in

  let include_user_file ctx file =
    get_file false (ctx.user_mk_path @ ctx.sys_mk_path) file
  in

  let include_sys_file ctx file =
    get_file true ctx.sys_mk_path file
  in

  let sys_cmds = 
    if sys_mk then
      include_sys_file ctx "sys.mk"
    else
      [] 
  in

  let cmds = get_file false [ Godi_file.filename "." ] main_file in
  eval
    ~pref_user_error ~include_user_file ~include_sys_file ctx (sys_cmds @ cmds)
;;


(**********************************************************************)
(* Executor                                                           *)
(**********************************************************************)

type exec_state =
    { mutable made_targets : SSet.t }
      (* Already made targets. Targets that are made unconditionally are
       * not added to the list:
       * - Targets with "!" operator
       * - Targets with "::" operator and no sources
       * - Phony targets
       *
       * Furthermore, .MADE rules aren't added, too.
       *)


let target_file ctx target =
  let target_fn = Godi_file.filename target in
  Godi_file.concat ctx.objdir target_fn
;;

let target_exists ctx target =
  Godi_file.file_exists (target_file ctx target)
;;


let target_mtime ctx target =
  Godi_file.file_mtime (target_file ctx target)
;;


let touch_target ctx target =
  let file = target_file ctx target in
  let ch = Godi_file.open_out_gen [Open_wronly; Open_creat] 0o666 file in
  close_out ch
;;


let rec find_value f l =
  (* Find the first element x of l for which f x does not raise Not_found *)
  match l with
    | [] -> raise Not_found
    | x :: l' -> (try f x with Not_found -> find_value f l')
;;


exception Skip ;;

let rec make_target exec_mode
		    exec_state
		    descendents    (* targets currently being made *)
                    ctx
		    level
		    target =

  let out = printmsg ctx.console in

  if ctx.trace then
    out ("*** Making [" ^ string_of_int level ^ "] " ^ target ^ "\n");

  (* Check whether [target] is up to date, and if not, create it. 
   * Return [true] if target has been created, else [false].
   *)
  let s_target = S.mk target in

  try
    (* If already made, do not make again. *)
    if SSet.mem s_target exec_state.made_targets then (
      if ctx.trace then out "    (Skipping because already made)\n";
      raise Skip;
    );
    
    (* Check for loops: *)
    if SSet.mem s_target descendents then (
      (* Output a warning and drop this target: *)
      out ("Cyclic dependency for target: " ^ target ^ "\n");
      raise Skip
    );

    (* Check whether there is an explicit rule: *)
    let chain =  (* A list of (rules) elements *)
      try
	let rules = H.find ctx.rules s_target in (* or Not_found *)
	(* If there aren't any commands in a rule, we check whether we
         * have implicit rules with commands
	 *)
	let rules' =
	  List.flatten
	    (List.map
	       (fun r ->
		  if r.rule_commands = [] && r.rule_file_type = `File &&
		    r.rule_type <> `Concat_dep
		  then
		    (* Check if the command can be taken from an implicit rule *)
		    try
		      let (imp_source, target_suffix, rules) =
			search_implicit_rule ctx r.rule_target in
		      (* Of course, we aren't interested in real chains here! *)
		      List.map
			(derive_from_implicit_rule target target_suffix
			   (imp_source :: r.rule_sources))
			rules
		    with
			Not_found ->
			  (* raise (Target_not_found target) -- CHECK *)
			  [r]
		  else
		    [r]
	       )
	       rules) in
	[ rules' ]
      with
	  Not_found ->  (* raised by H.find *)
	    (* Maybe the file already exists: *)
	    if target_exists ctx target then (
	      if ctx.trace then 
		out "    (File exists and no rule to check it)\n";
	      raise Skip;
	    );

	    (* Search for an implicit chain *)
	    ( try
		let imp_chain = search_implicit_chain ctx target in
		(* Or Not_found *)
		if ctx.trace then
		  out "    (Implicit rules applied)\n";
		List.map
		  (fun (imp_source, target_suffix, rules) ->
		     let derived_rules =
		       List.map 
			 (derive_from_implicit_rule 
			    target target_suffix [imp_source])
			 rules in
		     derived_rules
		  )
		  imp_chain
	      with
		  Not_found ->
		    raise (Target_not_found target)
	    )
    in
    
    (* Now execute the chain. If the chain fails, we must "roll back" its
     * transaction as a whole: All targets are deleted in turn, unless
     * they are declared as "precious".
     *
     * Even if successful, the intermediate files of the chain must be
     * deleted afterwards. (Otherwise the chain would not be found the
     * next time "make" is invoked.)
     *)
    let transaction = ref [] in
    let ood_flag = ref false in
    ( try
	List.iter
	  (fun rules ->
	     let chain_target = (List.hd rules).rule_target in
	     let is_intermediate = chain_target <> target in
	     let is_precious = 
	       ctx.global_precious_flag ||
	       List.mem chain_target ctx.precious_targets in
	     let file = target_file ctx chain_target in
	     if not (Godi_file.file_exists file) && not is_precious then
	       transaction := (is_intermediate, file) :: !transaction;
	     (* Only delete new files! *)

	     (* Execute the found rules in turn: *)
	     List.iter
	       (fun r ->
		  let u =
		    exec_rule_if_outdated 
		      exec_mode exec_state descendents ctx level r
		  in
		  ood_flag := !ood_flag || u;
	       )
	       rules
	  )
	  chain
      with
	  err ->   (* Note: This includes the case Sys.Break *)
	    List.iter 
	      (fun (_,file) -> try Godi_file.remove file with _ -> ()) 
	      !transaction;
	    raise err
    );
    (* Delete intermediate files: *)
    List.iter 
      (fun (is_im,file) -> 
	 if is_im then try Godi_file.remove file with _ -> ()) 
      !transaction;

    !ood_flag

  with
      Skip ->   (* no need for an update *)
	false
	  

and derive_from_implicit_rule target target_suffix sources r =
  { r with
      rule_target = target;
      rule_target_suffix = target_suffix;
      rule_sources = sources;
      rule_atts = [];
      rule_file_type = `File;
      (* an implicit rule cannot be phony! *)
  }

and search_implicit_rule ctx target =
  (* Search the implicit rule (source,rules), or raise Not_found. *)
  let chain =
    search_implicit_chain_in false ctx.suffixes ctx target in
  match chain with
    | [ chain_hd ] -> chain_hd
    | _ -> assert false   (* Other values are impossible *)

and search_implicit_chain ctx target =
  (* Search a chain of rules [ (source1,rules1); ... ] or raise Not_found *)
  search_implicit_chain_in true ctx.suffixes ctx target

and search_implicit_chain_in enable_chaining suffixes ctx target =
  (* First check if the target ends with a registered suffix: *)
  let target_suffixes =
    List.filter
      (fun suffix -> 
	 Filename.check_suffix target suffix) 
      suffixes in
  find_value
    (fun suffix ->
       let suffixes' = List.filter (fun s -> s <> suffix) suffixes in
       search_implicit_chain_in' enable_chaining suffixes' ctx target suffix
    )
    target_suffixes

and search_implicit_chain_in' enable_chaining suffixes' ctx target target_suffix =
  let target_base = 
    Filename.chop_suffix target target_suffix in
  (* Second search the source suffix, so we have a pair. First, we simply take
   * the first source suffix that has a corresponding file, or a 
   * corresponding explicit rule. Only if this does not succeed, we must
   * recurse.
   *)
  let source_suffix, chain_tl =
    try
      let s =
	List.find
	  (fun suffix ->
	     let have_implicit_rule =
	       H.mem ctx.rules (S.mk (suffix ^ target_suffix)) in
	     let source = target_base ^ suffix in
	     have_implicit_rule &&
	       (target_exists ctx source || H.mem ctx.rules (S.mk source))
	  )
	  suffixes' in
      (s, [])   (* No further elements in the chain *)
    with
	Not_found when enable_chaining ->
	  (* Try to construct a chain *)
	  find_value
	    (fun suffix ->
	       (* Note: Not_found is caught by [find_value] *)
	       let have_implicit_rule =
		 H.mem ctx.rules (S.mk (suffix ^ target_suffix)) in
	       if not have_implicit_rule then raise Not_found;
	       let source = target_base ^ suffix in
	       let chain_tl =
		 search_implicit_chain_in true suffixes' ctx source in
	         (* or Not_found *)
	       (suffix, chain_tl)
	    )
	    suffixes'
  in
  let source = target_base ^ source_suffix in
  let rules = H.find ctx.rules (S.mk(source_suffix ^ target_suffix)) in
  (source, target_suffix, rules) :: chain_tl

and exec_rule_if_outdated exec_model
		          exec_state
		          descendents
			  ctx
			  level
			  r =
  let out = printmsg ctx.console in

  (* If the rule is .MADE, don't recurse *)
  if r.rule_made_flag then (
    if ctx.trace then out "    (Target marked as .MADE)\n";
    false
  )
  else (
    try
      let descendents' =
	SSet.add (S.mk r.rule_target) descendents in
      
      (* Investigate the sources, and ensure they are all up to date *)
      let sources_results =
	List.map
	  (fun source ->
	     make_target exec_model exec_state descendents' ctx (level+1) source
	       (* may raise Target_not_found *)
	  )
	  r.rule_sources 
      in
      
      let source_outdated =
	List.exists (fun b -> b) sources_results in
      
      let sources_ood =
	List.flatten
	  (List.map2
	     (fun source result ->
		if result then [source] else [])
	     r.rule_sources
	     sources_results) in
      
      (* Check whether the target needs to be made *)
      let target_outdated =
	match (r.rule_file_type, r.rule_type) with
	  | (`Phony, _) ->
	      (* A .PHONY rule is always outdated *)
	      true
		
	  | (`Join, _) ->
	      (* A .JOIN rule is only outdated if a source is outdated *)
	      source_outdated
		
	  | (`File, `Dep) ->
	      (* A normal dependency is outdated if:
               * - a source is outdated, or
               * - the target does not exist, or
               * - the target is older than the newest existing source
               *)
	      source_outdated ||
		not (target_exists ctx r.rule_target) ||
		(let t_mtime = target_mtime ctx r.rule_target in
		 List.exists 
		   (fun mtime -> t_mtime < mtime)
		   (List.map (fun source ->
				try target_mtime ctx source
				with _ -> 0.0) r.rule_sources))
		
	  | (`File, `Forced_dep) ->
	      (* A forced dependency is always out of date *)
	      true
		
	  | (`File, `Concat_dep) ->
	      (* A :: dependency is outdated if:
               * - there are no sources, or
               * - a source is outdated, or
               * - the target does not exist, or
               * - the target is older than the newest existing source
               *)
	      (r.rule_sources = []) ||
		source_outdated ||
		not (target_exists ctx r.rule_target) ||
		(let t_mtime = target_mtime ctx r.rule_target in
		 List.exists 
		   (fun mtime -> t_mtime < mtime)
		   (List.map (fun source ->
				try target_mtime ctx source
				with _ -> 0.0) r.rule_sources))
      in
      
      (* If the target is outdated, create it: *)
      if target_outdated then (
	let magic = 
	  { magic_target = Some r.rule_target;
	    magic_target_suffix = Some r.rule_target_suffix;
	    magic_sources = r.rule_sources;
	    magic_sources_ood = sources_ood;
	    magic_exec_level = Some (level+1);
	  }
	in
	if ctx.trace then
	  out ("*** Executing [" ^ string_of_int level ^ "] " ^ r.rule_target ^ "\n");
	match exec_model with
	  | `Execute ->
	      List.iter (exec_command ctx r magic) r.rule_commands
	  | `Simulate_except ->
	      if r.rule_force_exec_flag then
		List.iter (exec_command ctx r magic) r.rule_commands
	      else
		List.iter (show_command ctx r magic) r.rule_commands
	  | `Simulate ->
	      List.iter (show_command ctx r magic) r.rule_commands
	  | `Test ->
	      ()
	  | `Fake ->
	      touch_target ctx r.rule_target
      );
      
      (* Remember our action in some cases: *)
      let remember =
	match (r.rule_file_type, r.rule_type) with
	  | (`Phony, _) -> false
	  | (_, `Forced_dep) -> false
	  | (_, `Concat_dep) -> r.rule_sources <> []
	  | (_, _) -> true in
      if remember then (
	exec_state.made_targets <- 
	  SSet.add (S.mk r.rule_target) exec_state.made_targets
      );
      
      target_outdated
    with
	Target_not_found _ when r.rule_optional_flag ->
	  if ctx.trace then
	    out ("*** Ignoring optional target: " ^ r.rule_target ^ "\n");
	  false   (* assume the target exists *)
  )

and expand_command ctx r magic (loc,value) =
  let s =
    expand ~magic_context:magic ctx loc value in
  let ignore_flag = ref false in
  let quiet_flag = ref false in
  let n = String.length s in
  let k = ref 0 in
  while !k < n && 
        (s.[ !k ] = '-' || s.[ !k ] = '@' || s.[ !k ] = ' ' || s.[ !k ] = '\t')
  do
    match s.[ !k ] with
	'-' -> ignore_flag := true; incr k
      | '@' -> quiet_flag := true; incr k
      | ' '
      | '\t' -> incr k
      | _   -> assert false
  done;
  let s' = String.sub s !k (n - !k) in
  (s', !ignore_flag, !quiet_flag)

and show_command ctx r magic (loc,value) =
  let out = printmsg ctx.console in
  let (cmd, ignore_flag, quiet_flag) = 
    expand_command ctx r magic (loc,value) in
  (* Ignore ignore_flag here: *)
  out (cmd ^ "\n")

and exec_command ctx r magic (loc,value) =
  let out = printmsg ctx.console in
  let (cmd_str, ignore_flag, quiet_flag) = 
    expand_command ctx r magic (loc,value) in
  let be_silent =
    not ctx.trace &&
    (quiet_flag || r.rule_silent_flag || ctx.global_silent_flag) in
  let ignore_failure =
    ignore_flag || r.rule_ignore_flag || ctx.global_ignore_flag in
  if not be_silent then
    out (cmd_str ^ "\n");
  let cmd =
    command ~magic_context:magic ctx cmd_str in
  let status = ctx.call_sh_cmd ctx cmd in
  match status with
    | Godi_exec.WEXITED 0 ->
	()
    | Godi_exec.WEXITED n ->
	if not ignore_failure then (
	  (* Note: No need to delete the target, this is done above *)
	  raise(Exec_error(loc, "Command returned with non-zero exit code"))
	)
    | Godi_exec.WSIGNALED _ ->
	if not ignore_failure then (
	  (* Note: No need to delete the target, this is done above *)
	  raise(Exec_error(loc, "Command terminated by signal"))
	)
    | Godi_exec.WSTOPPED _ ->
	assert false
;;


let execute ?(pref_user_error = false)
            ?(exec_mode = `Execute)
	    level
	    ctx =

  let out = printmsg ctx.console in

  if ctx.trace then
    out ("*** Entering [" ^ string_of_int level ^ "] " ^ 
	   Godi_file.slashname ctx.objdir ^ "\n");

  catch_user_error
    ~pref_user_error
    (fun () -> 
       let main_targets =
	 ( if H.mem ctx.rules (S.mk ".BEGIN") then [ ".BEGIN" ] else [] ) @
	   ( if ctx.final_targets = [] then
	       ( match ctx.default_target with
		   | None ->
		       raise(Exec_error(loc_unknown, "No target specified"))
		   | Some s ->
		       [s]
	       ) 
	     else
	       ctx.final_targets ) @
	   ( if H.mem ctx.rules (S.mk ".END") then [ ".END" ] else [] )
       in
       
       let exec_state =
	 { made_targets = SSet.empty } in
       
       let ood_list =
	 List.map
	   (fun target ->
	      make_target exec_mode exec_state SSet.empty ctx level target)
	   main_targets in
       
       let ood = List.exists (fun b -> b) ood_list in
       
       if exec_mode = `Test then
	 ( if ood then raise Out_of_date )
       else
	 ( if not ood then
	     let msg =
	       match main_targets with
		 | [] -> assert false
		 | [target] ->
		     "\"" ^ target ^ "\" is up to date\n"
		 | _ ->
		     String.concat ", "
		       (List.map (fun target -> "\"" ^ target ^ "\"") main_targets)
		     ^ " are up to date" in
	     out msg
	 );
    )
    ();
       
  if ctx.trace then 
    out ("*** Leaving [" ^ string_of_int level ^ "] " ^ 
	   Godi_file.slashname ctx.objdir ^ "\n");
;;

(**********************************************************************)
(* API *)

let expand ?pref_user_error ?(magic_context = no_magic) ?enable_env_check
           ctx loc v =
  expand ?pref_user_error ~magic_context ?enable_env_check ctx loc v

let expand_var ?pref_user_error ?(magic_context = no_magic) ?enable_env_check
           ctx loc s =
  expand_var ?pref_user_error ~magic_context ?enable_env_check ctx loc s

(**********************************************************************)
(* Notes about compatibility, limitations:
 *
 * BACKSLASHES:
 *
 * - In "bmake", the code to recognize comments is broken. A '#' preceded
 *   by a backslash is not a comment character, but what to do with a
 *   double-backslash followed by '#'?
 *   We do it right here: \\# is a backslash followed by a comment,
 *   but \\\# is a backslash followed by a '#'.
 *
 * - On the rhs of variable definitions, backslash is always
 *   considered as an escaping character. It depends on the context
 *   which characters can be escaped:
 *    - Delimiters (e.g. ${x\}x}) expands the value of the variable
 *      "x}x"
 *    - Other meta characters, e.g. * in glob expressions
 *    - Always: backslash, space, TAB, '#'
 *
 *   C-like backslash sequences (\n, \ddd etc) are not supported.
 *
 *   Of course, this treatment of backslash does not make sense inside
 *   shell commands.
 *
 * OTHER:
 *
 * - .if/.endif (and the other structured commands) must be properly nested
 *   with respect to .include, i.e. you cannot have the .if and .endif in
 *   different files
 *
 * - Comparisons with floats are slightly different:
 *   $(X) == $(Y) 
 *   In this case, godi_make performs always a string comparison, even
 *   if X and Y look like numbers. Use $(X) >= $(Y) && $(Y) <= $(X) to
 *   force a numeric comparison.
 *
 * - .for: the variable bound by .for is treated as a locally scoped
 *   variable. After .for, the variable has again the same value as
 *   before .for. In "bmake", however, the variable is reset to undef.
 *   (Same for the @-modifier.)
 *
 * - .PATH/VPATH is not supported. This feature is unsound.
 *
 * - Archives are not supported.
 *
 * - .USE, .USEBEFORE, .DEFAULT are not supported
 *
 * - .INTERRUPT is not supported
 *
 * - Some cases are errors whereas "bmake" only prints warnings:
 *   * When commands run by "!=" assignments exit with non-zero exit codes
 *   * Duplicate scripts
 *
 * - The file ".depend" is not automatically read, use .include
 *
 * - If an included file does not exist, it is tried to make. This is only
 *   done once, however, an automatism to remake Makefiles a la GNU make
 *   has not been implemented.
 *
 * - Extended regular expressions are not fully implemented (see Godi_match)
 *
 * - Assignment modifiers are not supported (e.g. ${x::=y}). Note that
 *   .for variables allow modifiers in Godi_make, so you don't need assignments
 *   to get around of any limitations of .for variables.
 *
 * - .undef: This undefines the variable _always_, even if there is
 *   an environment variable with the same name. In contrast to this, 
 *   "bmake" interprets .undef as "make the environment variable
 *   again visible".
 *
 * - If commands fail, the target is deleted. (Like .DELETE_ON_ERROR in
 *   GNU make.)
 *
 * - Parallel command execution is not supported
 *)

(* CHECK:
 * X=V\
 *)
